#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import math
import geoclide as gc
from geoclide.shapes import get_intersect_dataset
import os

ROOTPATH = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))


def test_triangle_intersection():
    p0 = gc.Point(-1., -1., 0.)
    p1 = gc.Point(1., -1., 0.)
    p2 = gc.Point(1., 1., 0.)

    tri = gc.Triangle(p0, p1, p2)
    assert (tri.area() == 2.)

    ray = gc.Ray(o=gc.Point(0., 0., 1.), d=gc.normalize(gc.Vector(0.999,0.999,-1.)))

    ds_v2 = tri.intersect_v2(ray)
    ds_v3 = tri.intersect_v3(ray)

    assert (ds_v2['is_intersection'].item()), 'Problem with v2 intersection test'
    assert (ds_v3['is_intersection'].item()), 'Problem with v3 intersection test'
    assert (np.isclose(1.73089629960896, ds_v2['thit'].item(), 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(1.73089629960896, ds_v3['thit'].item(), 0., 1e-14)), 'Problem with v3 intersection test'
    assert (gc.Normal(ds_v2['nhit'].values) == gc.Normal(0., 0., 1.)), 'Problem with v2 intersection test'
    assert (gc.Normal(ds_v3['nhit'].values) == gc.Normal(0., 0., 1.)), 'Problem with v3 intersection test'

    p3 = gc.Point(0.999, 0.999, 0.)

    assert (np.isclose(ds_v2['phit'].values[0], p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v2['phit'].values[1], p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v2['phit'].values[2], p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[0], p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[1], p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[2], p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    # Bellow the ray cannot reach the triangle
    ray = gc.Ray(o=gc.Point(0., 0., 1.), d=gc.normalize(gc.Vector(0.999,0.999,-1.)), maxt=1.7)

    ds_v2 = tri.intersect_v2(ray)
    ds_v3 = tri.intersect_v3(ray)
    
    assert (ds_v2['is_intersection'].item() is False), 'Problem with v2 intersection test'
    assert (ds_v3['is_intersection'].item() is False), 'Problem with v3 intersection test'

def test_triangle_transform():
    p0 = gc.Point(-0.5, 0.5, 0.)
    p1 = gc.Point(0.5, 0.5, 0.)
    p2 = gc.Point(0.5, -0.5, 0.)

    oTw = gc.get_translate_tf(gc.Vector(10., 0., 5.)) * gc.get_rotateY_tf(45.)
    tri = gc.Triangle(p0, p1, p2, oTw=oTw)
    assert (tri.area() == 0.5)

    ray = gc.Ray(o=gc.Point(0., 0., 4.8), d=gc.normalize(gc.Vector(1.,0.,0.)))

    ds_v2 = tri.intersect(ray, method='v2')
    ds_v3 = tri.intersect(ray, method='v3')

    assert (ds_v2['is_intersection'].item()), 'Problem with v2 intersection test'
    assert (ds_v3['is_intersection'].item()), 'Problem with v3 intersection test'
    assert (np.isclose(10.2, ds_v2['thit'].item(), 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(10.2, ds_v3['thit'].item(), 0., 1e-14)), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., ds_v2['nhit'].values[0], 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (ds_v2['nhit'].values[1] == 0.), 'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., ds_v2['nhit'].values[2], 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., ds_v3['nhit'].values[0], 0., 1e-13)), \
        'Problem with v3 intersection test'
    assert (ds_v3['nhit'].values[1] == 0.), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., ds_v3['nhit'].values[2], 0., 1e-13)), \
        'Problem with v3 intersection test'

    p3 = gc.Point(10.2, 0., 4.8)

    assert (np.isclose(ds_v2['phit'].values[0], p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v2['phit'].values[1], p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v2['phit'].values[2], p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[0], p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[1], p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(ds_v3['phit'].values[2], p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    assert (tri.is_intersection(ray, method='v2')), 'Problem with v2 is_intersection test'
    assert (tri.is_intersection(ray, method='v3')), 'Problem with v3 is_intersection test'

    thitv2_, is_intv2_ = tri.is_intersection_t(ray, method='v2')
    thitv3_, is_intv3_ = tri.is_intersection_t(ray, method='v3')
    assert (is_intv2_)
    assert (is_intv3_)
    assert (ds_v2['thit'].item() == thitv2_)
    assert (ds_v3['thit'].item()  == thitv3_)

def test_triangle_mesh():
    # list of vertices
    vertices = np.array([ [-0.5, -0.5, 0.],                  # v0
                          [0.5, -0.5, 0.],                   # v1
                          [-0.5, 0.5, 0.],                   # v2
                          [0.5, 0.5, 0.]], dtype=np.float64) # v3

    faces = np.array([[0, 1, 2],                   # vertices index of T0
                      [2, 3, 1]], dtype=np.int32)  # vertices index of T1

    oTw = gc.get_translate_tf(gc.Vector(10., 0., 5.)) * gc.get_rotateY_tf(45.)
    tri_mesh = gc.TriangleMesh(vertices=vertices, faces=faces, oTw=oTw)
    assert (tri_mesh.area() == 1.)

    ray = gc.Ray(o=gc.Point(0., 0., 4.8), d=gc.normalize(gc.Vector(1.,0.,0.)))

    ds_v2 = gc.calc_intersection(tri_mesh, ray, method='v2')
    ds_v3 = gc.calc_intersection(tri_mesh, ray, method='v3')
    thitv2 = ds_v2['thit'].values
    nhitv2 = gc.Normal(ds_v2['nhit'].values)
    is_intv2 = bool(ds_v2['is_intersection'].values)
    thitv3 = ds_v3['thit'].values
    nhitv3 = gc.Normal(ds_v3['nhit'].values)
    is_intv3 = bool(ds_v3['is_intersection'].values)

    assert (is_intv2 is True), 'Problem with v2 intersection test'
    assert (is_intv3 is True), 'Problem with v3 intersection test'
    assert (np.isclose(10.2, thitv2, 0., 1e-14)), 'Problem with v2 intersection test'
    assert (np.isclose(10.2, thitv3, 0., 1e-14)), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., nhitv2.x, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (nhitv2.y == 0.), 'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., nhitv2.z, 0., 1e-13)), \
        'Problem with v2 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., nhitv3.x, 0., 1e-13)), \
        'Problem with v3 intersection test'
    assert (nhitv3.y == 0.), 'Problem with v3 intersection test'
    assert (np.isclose(-math.sqrt(2.)/2., nhitv3.z, 0., 1e-13)), \
        'Problem with v3 intersection test'

    p3 = gc.Point(10.2, 0., 4.8)
    phitv2 = gc.Point(ds_v2['phit'].values)
    phitv3 = gc.Point(ds_v3['phit'].values)

    assert (np.isclose(phitv2.x, p3.x, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(phitv2.y, p3.y, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(phitv2.z, p3.z, 0., 1e-15)), 'Problem with v2 triangle intersection test'
    assert (np.isclose(phitv3.x, p3.x, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(phitv3.y, p3.y, 0., 1e-15)), 'Problem with v3 triangle intersection test'
    assert (np.isclose(phitv3.z, p3.z, 0., 1e-15)), 'Problem with v3 triangle intersection test'

    assert (tri_mesh.is_intersection(ray, method='v2')), 'Problem with v2 is_intersection test'
    assert (tri_mesh.is_intersection(ray, method='v3')), 'Problem with v3 is_intersection test'


def test_read_gcnc_trianglemesh():
    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/sphere_r1p5_resth18_resph36.gcnc')
    msh = gc.Sphere(1.5).to_trianglemesh(reso_theta=18, reso_phi=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/sphere_r1p5_resth18_resph18_phimax180.gcnc')
    msh = gc.Sphere(1.5, phi_max=180.).to_trianglemesh(reso_theta=18, reso_phi=18)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/oblate_rxy1p5_rz1p2_resth18_resph36.gcnc')
    msh = gc.Spheroid(1.5, 1.2).to_trianglemesh(reso_theta=18, reso_phi=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/prolate_rxy1p5_rz3_resth18_resph36.gcnc')
    msh = gc.Spheroid(1.5, 3.).to_trianglemesh(reso_theta=18, reso_phi=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/disk_r1_reso36.gcnc')
    msh = gc.Disk(1.).to_trianglemesh(reso=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/disk_r1_zh5_reso36.gcnc')
    msh = gc.Disk(1., z_height=5.).to_trianglemesh(reso=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/annulus_r1_ir0p5_reso36.gcnc')
    msh = gc.Disk(1., inner_radius=0.5).to_trianglemesh(reso=36)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))

    msh_read = gc.read_trianglemesh(ROOTPATH + '/geoclide/tests/data/annulus_partial_r1_ir0p5_reso27_phimax270.gcnc')
    msh = gc.Disk(1., inner_radius=0.5, phi_max=270).to_trianglemesh(reso=27)
    assert (np.all(np.isclose(msh_read.vertices, msh.vertices, 0., 1e-15)))
    assert (np.all(msh_read.faces == msh.faces))


def test_trianglemesh_to_dataset():
    msh1 = gc.Sphere(1.5).to_trianglemesh(reso_theta=18, reso_phi=36)
    msh2 = msh1.to_dataset()
    assert (np.all(np.isclose(msh1.vertices, msh2['vertices'].values[0,:,:], 0., 1e-15)))
    assert (np.all(msh1.faces == msh2['faces'].values[0,:,:]))


def test_trianglemesh_numpy():
    msh = gc.Spheroid(radius_xy=1., radius_z=3).to_trianglemesh(reso_theta=45, reso_phi=90)
    r0 = gc.Ray(gc.Point(5., -0.2, 2.5), gc.Vector(-1., 0., 0.))
    ds_v2 = msh.intersect(r0, 'v2', use_loop=True)
    ds_v2f = msh.intersect(r0, 'v2', use_loop=False)
    ds_v3f = msh.intersect(r0, 'v3') # by default use_loop=False
    assert (ds_v2['is_intersection'].item() == ds_v2f['is_intersection'].item())
    assert (ds_v2['is_intersection'].item() == ds_v3f['is_intersection'].item())
    assert (np.isclose(ds_v2['thit'].item(), ds_v2f['thit'].item(), 0., 1e-15))
    assert (np.isclose(ds_v2['thit'].item(), ds_v3f['thit'].item(), 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[0], ds_v2f['phit'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[1], ds_v2f['phit'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[2], ds_v2f['phit'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[0], ds_v3f['phit'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[1], ds_v3f['phit'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['phit'].values[2], ds_v3f['phit'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[0], ds_v2f['dpdu'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[1], ds_v2f['dpdu'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[2], ds_v2f['dpdu'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[0], ds_v3f['dpdu'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[1], ds_v3f['dpdu'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdu'].values[2], ds_v3f['dpdu'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[0], ds_v2f['dpdv'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[1], ds_v2f['dpdv'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[2], ds_v2f['dpdv'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[0], ds_v3f['dpdv'].values[0], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[1], ds_v3f['dpdv'].values[1], 0., 1e-15))
    assert (np.isclose(ds_v2['dpdv'].values[2], ds_v3f['dpdv'].values[2], 0., 1e-15))
    assert (np.isclose(ds_v2['u'].item(), ds_v2f['u'].item(), 0., 1e-15))
    assert (np.isclose(ds_v2['u'].item(), ds_v3f['u'].item(), 0., 1e-15))
    assert (np.isclose(ds_v2['v'].item(), ds_v2f['v'].item(), 0., 1e-15))
    assert (np.isclose(ds_v2['v'].item(), ds_v3f['v'].item(), 0., 1e-15))


def test_triangle_2d_arr1():
    msh = gc.Sphere(1.).to_trianglemesh(reso_theta=5, reso_phi=5)
    x_, y_, z_ = np.meshgrid(np.linspace(-0.4, 0.4, 4, np.float64),
                            np.linspace(-0.4, 0.4, 4, np.float64),
                            2., indexing='ij')
    o_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    nrays = o_set_arr.shape[0]
    nobj = msh.ntriangles
    d_set_arr = np.zeros_like(o_set_arr)
    d_set_arr[:,0] = 0.
    d_set_arr[:,1] = 0.
    d_set_arr[:,2] = -1.
    o_set = gc.Point(o_set_arr)
    d_set = gc.Vector(d_set_arr)
    
    p0 = gc.Point(msh.vertices[msh.faces[:,0],:])
    p1 = gc.Point(msh.vertices[msh.faces[:,1],:])
    p2 = gc.Point(msh.vertices[msh.faces[:,2],:])
    triangles = gc.Triangle(p0, p1, p2)
    r_set = gc.Ray(o_set, d_set)
    ds_v2 = triangles.intersect_v2(r_set, diag_calc=False)
    ds_v3 = triangles.intersect_v3(r_set, diag_calc=False)

    t_v2, is_int_v2 = triangles.is_intersection_v2_t(r_set, diag_calc=False)
    t_v3, is_int_v3 = triangles.is_intersection_v3_t(r_set, diag_calc=False)
    assert (np.array_equal(ds_v2['is_intersection'].values, is_int_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['is_intersection'].values, is_int_v3, equal_nan=True))
    assert (np.array_equal(ds_v2['thit'].values, t_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, t_v3, equal_nan=True))

    is_int_2d = np.full((nobj, nrays), True, dtype=bool)
    t_2d = np.zeros((nobj, nrays), dtype=np.float64)
    p_2d = np.zeros((nobj, nrays,3), dtype=np.float64)
    n_2d = np.zeros_like(p_2d)
    dpdu_2d = np.zeros_like(p_2d)
    dpdv_2d = np.zeros_like(p_2d)
    u_2d = np.zeros_like(t_2d)
    v_2d = np.zeros_like(t_2d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for itri in range(0, nobj):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        for ir in range (0, nrays):
            res_sca = triangle.intersect_v2(list_rays[ir], ds_output=False)
            if res_sca[2] is not None:
                ds_sca = get_intersect_dataset(*res_sca)
                is_int_2d[itri,ir] = ds_sca['is_intersection'].values
                t_2d[itri,ir] = ds_sca['thit'].values
                p_2d[itri,ir,:] = ds_sca['phit'].values
                n_2d[itri,ir,:] = ds_sca['nhit'].values
                dpdu_2d[itri,ir,:] = ds_sca['dpdu'].values
                dpdv_2d[itri,ir,:] = ds_sca['dpdv'].values
                u_2d[itri,ir] = ds_sca['u'].values
                v_2d[itri,ir] = ds_sca['v'].values
            else:
                is_int_2d[itri,ir] = res_sca[3]
                t_2d[itri,ir] = res_sca[2]
                p_2d[itri,ir,:] = None
                n_2d[itri,ir,:] = None
                dpdu_2d[itri,ir,:] = None
                dpdv_2d[itri,ir,:] = None
                u_2d[itri,ir] = None
                v_2d[itri,ir] = None

    assert (np.array_equal(ds_v2['thit'].values, t_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['phit'].values, p_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['nhit'].values, n_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['u'].values, u_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['v'].values, v_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdu'].values, dpdu_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdv'].values, dpdv_2d, equal_nan=True))

    assert (np.allclose(ds_v3['thit'].values, t_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['phit'].values, p_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['nhit'].values, n_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['u'].values, u_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['v'].values, v_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['dpdu'].values, dpdu_2d, 0., 1e-15, equal_nan=True))
    assert (np.allclose(ds_v3['dpdv'].values, dpdv_2d, 0., 1e-15, equal_nan=True))

    is_int_2d = np.full((nobj, nrays), True, dtype=bool)
    t_2d = np.zeros((nobj, nrays), dtype=np.float64)
    p_2d = np.zeros((nobj, nrays,3), dtype=np.float64)
    n_2d = np.zeros_like(p_2d)
    dpdu_2d = np.zeros_like(p_2d)
    dpdv_2d = np.zeros_like(p_2d)
    u_2d = np.zeros_like(t_2d)
    v_2d = np.zeros_like(t_2d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for itri in range(0, nobj):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        for ir in range (0, nrays):
            res_sca = triangle.intersect_v3(list_rays[ir], ds_output=False)
            if res_sca[2] is not None:
                ds_sca = get_intersect_dataset(*res_sca)
                is_int_2d[itri,ir] = ds_sca['is_intersection'].values
                t_2d[itri,ir] = ds_sca['thit'].values
                p_2d[itri,ir,:] = ds_sca['phit'].values
                n_2d[itri,ir,:] = ds_sca['nhit'].values
                dpdu_2d[itri,ir,:] = ds_sca['dpdu'].values
                dpdv_2d[itri,ir,:] = ds_sca['dpdv'].values
                u_2d[itri,ir] = ds_sca['u'].values
                v_2d[itri,ir] = ds_sca['v'].values
            else:
                is_int_2d[itri,ir] = res_sca[3]
                t_2d[itri,ir] = res_sca[2]
                p_2d[itri,ir,:] = None
                n_2d[itri,ir,:] = None
                dpdu_2d[itri,ir,:] = None
                dpdv_2d[itri,ir,:] = None
                u_2d[itri,ir] = None
                v_2d[itri,ir] = None

    assert (np.array_equal(ds_v3['thit'].values, t_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['phit'].values, p_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['nhit'].values, n_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['u'].values, u_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['v'].values, v_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdu'].values, dpdu_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdv'].values, dpdv_2d, equal_nan=True))

    msh_ds = msh.intersect(r_set, use_loop=True)
    msh_dsn = msh.intersect(r_set, use_loop=False)
    assert (np.array_equal(msh_ds['thit'].values, msh_dsn['thit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['phit'].values, msh_dsn['phit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['nhit'].values, msh_dsn['nhit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['u'].values, msh_dsn['u'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['v'].values, msh_dsn['v'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdu'].values, msh_dsn['dpdu'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdv'].values, msh_dsn['dpdv'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(msh_ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))

    dimx=2
    dimy=2
    msh = gc.Sphere(4, oTw=gc.get_translate_tf(gc.Vector(0., 3.5, 10.))).to_trianglemesh(9,18)
    nx = 11
    ny = 11
    n_samples = nx*ny
    p_set_arr = np.zeros((n_samples,3), dtype=np.float64)
    hdim_x = dimx*0.5
    hdim_y = dimy*0.5
    deltax = (dimx/nx) * 0.5
    deltay = (dimy/ny) * 0.5
    x_, y_, z_ = np.meshgrid(np.linspace(-hdim_x+deltax, hdim_x-deltax, nx, np.float64),
                            np.linspace(-hdim_y+deltay, hdim_y-deltay, ny, np.float64),
                            0., indexing='ij')
    p_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    p_set = gc.Point(p_set_arr)
    d_set_arr = np.zeros_like(p_set_arr)
    d_set_arr[:,2] = 1.
    d_set = gc.Vector(d_set_arr)
    r_set = gc.Ray(p_set, d_set)
    ds = msh.intersect(r_set, use_loop=False)
    dsl = msh.intersect(r_set, use_loop=True)
    assert (np.array_equal(ds['thit'].values, dsl['thit'].values, equal_nan=True))
    assert (np.array_equal(ds['phit'].values, dsl['phit'].values, equal_nan=True))
    assert (np.array_equal(ds['nhit'].values, dsl['nhit'].values, equal_nan=True))
    assert (np.array_equal(ds['u'].values, dsl['u'].values, equal_nan=True))
    assert (np.array_equal(ds['v'].values, dsl['v'].values, equal_nan=True))
    assert (np.array_equal(ds['dpdu'].values, dsl['dpdu'].values, equal_nan=True))
    assert (np.array_equal(ds['dpdv'].values, dsl['dpdv'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r_set, use_loop=True)
    assert (np.array_equal(ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r_set, use_loop=True)
    assert (np.array_equal(ds['is_intersection'].values, is_int, equal_nan=True))


def test_triangle_2d_arr2():
    msh = gc.Sphere(1.).to_trianglemesh(reso_theta=4, reso_phi=4)
    x_, y_, z_ = np.meshgrid(np.linspace(-0.4, 0.4, 5, np.float64),
                            np.linspace(-0.4, 0.4, 5, np.float64),
                            2., indexing='ij')
    o_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    nrays = o_set_arr.shape[0]
    nobj = msh.ntriangles
    d_set_arr = np.zeros_like(o_set_arr)
    d_set_arr[:,0] = 0.
    d_set_arr[:,1] = 0.
    d_set_arr[:,2] = -1.
    o_set = gc.Point(o_set_arr)
    d_set = gc.Vector(d_set_arr)
    
    p0 = gc.Point(msh.vertices[msh.faces[:,0],:])
    p1 = gc.Point(msh.vertices[msh.faces[:,1],:])
    p2 = gc.Point(msh.vertices[msh.faces[:,2],:])
    triangles = gc.Triangle(p0, p1, p2)
    r_set = gc.Ray(o_set, d_set)
    ds_v2 = triangles.intersect_v2(r_set, diag_calc=False)

    # In v3 -> more robust test! even 1e-17 instead of 0. can lead to a failed test
    ds_v3 = triangles.intersect_v3(r_set, diag_calc=False)

    t_v2, is_int_v2 = triangles.is_intersection_v2_t(r_set, diag_calc=False)
    t_v3, is_int_v3 = triangles.is_intersection_v3_t(r_set, diag_calc=False)
    assert (np.array_equal(ds_v2['is_intersection'].values, is_int_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['is_intersection'].values, is_int_v3, equal_nan=True))
    assert (np.array_equal(ds_v2['thit'].values, t_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, t_v3, equal_nan=True))

    is_int_2d = np.full((nobj, nrays), True, dtype=bool)
    t_2d = np.zeros((nobj, nrays), dtype=np.float64)
    p_2d = np.zeros((nobj, nrays,3), dtype=np.float64)
    n_2d = np.zeros_like(p_2d)
    dpdu_2d = np.zeros_like(p_2d)
    dpdv_2d = np.zeros_like(p_2d)
    u_2d = np.zeros_like(t_2d)
    v_2d = np.zeros_like(t_2d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for itri in range(0, nobj):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        for ir in range (0, nrays):
            res_sca = triangle.intersect_v2(list_rays[ir], ds_output=False)
            if res_sca[2] is not None:
                ds_sca = get_intersect_dataset(*res_sca)
                is_int_2d[itri,ir] = ds_sca['is_intersection'].values
                t_2d[itri,ir] = ds_sca['thit'].values
                p_2d[itri,ir,:] = ds_sca['phit'].values
                n_2d[itri,ir,:] = ds_sca['nhit'].values
                dpdu_2d[itri,ir,:] = ds_sca['dpdu'].values
                dpdv_2d[itri,ir,:] = ds_sca['dpdv'].values
                u_2d[itri,ir] = ds_sca['u'].values
                v_2d[itri,ir] = ds_sca['v'].values
            else:
                is_int_2d[itri,ir] = False
                t_2d[itri,ir] = None
                p_2d[itri,ir,:] = None
                n_2d[itri,ir,:] = None
                dpdu_2d[itri,ir,:] = None
                dpdv_2d[itri,ir,:] = None
                u_2d[itri,ir] = None
                v_2d[itri,ir] = None

    assert (np.array_equal(ds_v2['thit'].values, t_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['phit'].values, p_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['nhit'].values, n_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['u'].values, u_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['v'].values, v_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdu'].values, dpdu_2d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdv'].values, dpdv_2d, equal_nan=True))

    is_int_2d = np.full((nobj, nrays), True, dtype=bool)
    t_2d = np.zeros((nobj, nrays), dtype=np.float64)
    p_2d = np.zeros((nobj, nrays,3), dtype=np.float64)
    n_2d = np.zeros_like(p_2d)
    dpdu_2d = np.zeros_like(p_2d)
    dpdv_2d = np.zeros_like(p_2d)
    u_2d = np.zeros_like(t_2d)
    v_2d = np.zeros_like(t_2d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for itri in range(0, nobj):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        for ir in range (0, nrays):
            res_sca = triangle.intersect_v3(list_rays[ir], ds_output=False)
            if res_sca[2] is not None:
                ds_sca = get_intersect_dataset(*res_sca)
                is_int_2d[itri,ir] = ds_sca['is_intersection'].values
                t_2d[itri,ir] = ds_sca['thit'].values
                p_2d[itri,ir,:] = ds_sca['phit'].values
                n_2d[itri,ir,:] = ds_sca['nhit'].values
                dpdu_2d[itri,ir,:] = ds_sca['dpdu'].values
                dpdv_2d[itri,ir,:] = ds_sca['dpdv'].values
                u_2d[itri,ir] = ds_sca['u'].values
                v_2d[itri,ir] = ds_sca['v'].values
            else:
                is_int_2d[itri,ir] = False
                t_2d[itri,ir] = None
                p_2d[itri,ir,:] = None
                n_2d[itri,ir,:] = None
                dpdu_2d[itri,ir,:] = None
                dpdv_2d[itri,ir,:] = None
                u_2d[itri,ir] = None
                v_2d[itri,ir] = None

    assert (np.array_equal(ds_v3['thit'].values, t_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['phit'].values, p_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['nhit'].values, n_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['u'].values, u_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['v'].values, v_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdu'].values, dpdu_2d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdv'].values, dpdv_2d, equal_nan=True))

    msh_ds = msh.intersect(r_set, use_loop=True)
    msh_dsn = msh.intersect(r_set, use_loop=False)
    assert (np.array_equal(msh_ds['thit'].values, msh_dsn['thit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['phit'].values, msh_dsn['phit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['nhit'].values, msh_dsn['nhit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['u'].values, msh_dsn['u'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['v'].values, msh_dsn['v'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdu'].values, msh_dsn['dpdu'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdv'].values, msh_dsn['dpdv'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(msh_ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))


def test_triangle_1d_arr1():
    msh = gc.Sphere(1.).to_trianglemesh(reso_theta=5, reso_phi=5)
    o = gc.Point(-0.4, 0.4, 2.)
    d = gc.Vector(0., 0., -1.)
    r0 = gc.Ray(o, d)

    p0 = gc.Point(msh.vertices[msh.faces[:,0],:])
    p1 = gc.Point(msh.vertices[msh.faces[:,1],:])
    p2 = gc.Point(msh.vertices[msh.faces[:,2],:])
    triangles = gc.Triangle(p0, p1, p2)
    ds_v2 = triangles.intersect_v2(r0, diag_calc=False)
    ds_v3 = triangles.intersect_v3(r0, diag_calc=False)

    t_v2, is_int_v2 = triangles.is_intersection_v2_t(r0, diag_calc=False)
    t_v3, is_int_v3 = triangles.is_intersection_v3_t(r0, diag_calc=False)
    assert (np.array_equal(ds_v2['is_intersection'].values, is_int_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['is_intersection'].values, is_int_v3, equal_nan=True))
    assert (np.array_equal(ds_v2['thit'].values, t_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, t_v3, equal_nan=True))

    is_int_1d = np.full((msh.ntriangles), True, dtype=bool)
    t_1d = np.zeros((msh.ntriangles), dtype=np.float64)
    p_1d = np.zeros((msh.ntriangles,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)

    for itri in range(0, msh.ntriangles):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        res_sca = triangle.intersect_v2(r0, ds_output=False)
        if res_sca[2] is not None:
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[itri] = ds_sca['is_intersection'].values
            t_1d[itri] = ds_sca['thit'].values
            p_1d[itri,:] = ds_sca['phit'].values
            n_1d[itri,:] = ds_sca['nhit'].values
            dpdu_1d[itri,:] = ds_sca['dpdu'].values
            dpdv_1d[itri,:] = ds_sca['dpdv'].values
            u_1d[itri] = ds_sca['u'].values
            v_1d[itri] = ds_sca['v'].values
        else:
            is_int_1d[itri] = False
            t_1d[itri] = None
            p_1d[itri,:] = None
            n_1d[itri,:] = None
            dpdu_1d[itri,:] = None
            dpdv_1d[itri,:] = None
            u_1d[itri] = None
            v_1d[itri] = None

    assert (np.array_equal(ds_v2['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdv'].values, dpdv_1d, equal_nan=True))

    for itri in range(0, msh.ntriangles):
        p0 = gc.Point(msh.vertices[msh.faces[itri,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[itri,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[itri,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        res_sca = triangle.intersect_v3(r0, ds_output=False)
        if res_sca[2] is not None:
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[itri] = ds_sca['is_intersection'].values
            t_1d[itri] = ds_sca['thit'].values
            p_1d[itri,:] = ds_sca['phit'].values
            n_1d[itri,:] = ds_sca['nhit'].values
            dpdu_1d[itri,:] = ds_sca['dpdu'].values
            dpdv_1d[itri,:] = ds_sca['dpdv'].values
            u_1d[itri] = ds_sca['u'].values
            v_1d[itri] = ds_sca['v'].values
        else:
            is_int_1d[itri] = False
            t_1d[itri] = None
            p_1d[itri,:] = None
            n_1d[itri,:] = None
            dpdu_1d[itri,:] = None
            dpdv_1d[itri,:] = None
            u_1d[itri] = None
            v_1d[itri] = None

    assert (np.array_equal(ds_v3['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdv'].values, dpdv_1d, equal_nan=True))

    msh_ds = msh.intersect(r0, use_loop=True)
    msh_dsn = msh.intersect(r0, use_loop=False)
    assert (np.array_equal(msh_ds['thit'].values, msh_dsn['thit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['phit'].values, msh_dsn['phit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['nhit'].values, msh_dsn['nhit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['u'].values, msh_dsn['u'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['v'].values, msh_dsn['v'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdu'].values, msh_dsn['dpdu'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdv'].values, msh_dsn['dpdv'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r0, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(msh_ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r0, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))

def test_triangle_1d_arr2():

    x_, y_, z_ = np.meshgrid(np.linspace(-0.4, 0.4, 5, np.float64),
                         np.linspace(-0.4, 0.4, 5, np.float64),
                         2., indexing='ij')
    o_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    nrays = o_set_arr.shape[0]
    d_set_arr = np.zeros_like(o_set_arr)
    d_set_arr[:,0] = 0.
    d_set_arr[:,1] = 0.
    d_set_arr[:,2] = -1.
    o_set = gc.Point(o_set_arr)
    d_set = gc.Vector(d_set_arr)
    r_set = gc.Ray(o_set, d_set)

    p0 = gc.Point(np.array([0., 0., 1]))
    p1 = gc.Point(np.array([0.70710678, 0., 0.70710678]))
    p2 = gc.Point(np.array([0.21850801, 0.67249851, 0.70710678]))
    triangle = gc.Triangle(p0, p1, p2)

    ds_v2 = triangle.intersect_v2(r_set, diag_calc=False)
    ds_v3 = triangle.intersect_v3(r_set, diag_calc=False)

    t_v2, is_int_v2 = triangle.is_intersection_v2_t(r_set, diag_calc=False)
    t_v3, is_int_v3 = triangle.is_intersection_v3_t(r_set, diag_calc=False)
    assert (np.array_equal(ds_v2['is_intersection'].values, is_int_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['is_intersection'].values, is_int_v3, equal_nan=True))
    assert (np.array_equal(ds_v2['thit'].values, t_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, t_v3, equal_nan=True))

    is_int_1d = np.full((nrays), True, dtype=bool)
    t_1d = np.zeros((nrays), dtype=np.float64)
    p_1d = np.zeros((nrays,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for ir in range (0, nrays):
        res_sca = triangle.intersect_v2(list_rays[ir], ds_output=False)
        if (res_sca[2] is not None):
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[ir] = ds_sca['is_intersection'].values
            t_1d[ir] = ds_sca['thit'].values
            p_1d[ir,:] = ds_sca['phit'].values
            n_1d[ir,:] = ds_sca['nhit'].values
            dpdu_1d[ir,:] = ds_sca['dpdu'].values
            dpdv_1d[ir,:] = ds_sca['dpdv'].values
            u_1d[ir] = ds_sca['u'].values
            v_1d[ir] = ds_sca['v'].values
        else:
            is_int_1d[ir] = False
            t_1d[ir] = None
            p_1d[ir,:] = None
            n_1d[ir,:] = None
            dpdu_1d[ir,:] = None
            dpdv_1d[ir,:] = None
            u_1d[ir] = None
            v_1d[ir] = None

    assert (np.array_equal(ds_v2['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdv'].values, dpdv_1d, equal_nan=True))

    is_int_1d = np.full((nrays), True, dtype=bool)
    t_1d = np.zeros((nrays), dtype=np.float64)
    p_1d = np.zeros((nrays,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for ir in range (0, nrays):
        res_sca = triangle.intersect_v3(list_rays[ir], ds_output=False)
        if (res_sca[2] is not None):
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[ir] = ds_sca['is_intersection'].values
            t_1d[ir] = ds_sca['thit'].values
            p_1d[ir,:] = ds_sca['phit'].values
            n_1d[ir,:] = ds_sca['nhit'].values
            dpdu_1d[ir,:] = ds_sca['dpdu'].values
            dpdv_1d[ir,:] = ds_sca['dpdv'].values
            u_1d[ir] = ds_sca['u'].values
            v_1d[ir] = ds_sca['v'].values
        else:
            is_int_1d[ir] = False
            t_1d[ir] = None
            p_1d[ir,:] = None
            n_1d[ir,:] = None
            dpdu_1d[ir,:] = None
            dpdv_1d[ir,:] = None
            u_1d[ir] = None
            v_1d[ir] = None

    assert (np.array_equal(ds_v3['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdv'].values, dpdv_1d, equal_nan=True))

    vertices = np.array([[0.        , 0.        , 1         ],
                         [0.70710678, 0.        , 0.70710678],
                         [0.21850801, 0.67249851, 0.70710678]])
    faces = np.array([[0, 1, 2]])
    msh = gc.TriangleMesh(vertices=vertices, faces=faces)
    msh_ds = msh.intersect(r_set, use_loop=True)
    msh_dsn = msh.intersect(r_set, use_loop=False)
    assert (np.array_equal(msh_ds['thit'].values, msh_dsn['thit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['phit'].values, msh_dsn['phit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['nhit'].values, msh_dsn['nhit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['u'].values, msh_dsn['u'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['v'].values, msh_dsn['v'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdu'].values, msh_dsn['dpdu'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdv'].values, msh_dsn['dpdv'].values, equal_nan=True))
    # Particular case where we should get the same values bellow (because only 1 triangle)
    assert (np.array_equal(ds_v3['is_intersection'].values,
                           msh_ds['is_intersection'].values, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, msh_ds['thit'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(msh_ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r_set, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))


def test_triangle_1d_arr3():
    msh = gc.Sphere(1.).to_trianglemesh(reso_theta=4, reso_phi=4)
    x_, y_, z_ = np.meshgrid(np.linspace(0.0, 0.6, 4, np.float64),
                         np.linspace(0., 0.4, 4, np.float64),
                         2., indexing='ij')
    o_set_arr = np.vstack((x_.ravel(), y_.ravel(), z_.ravel())).T
    nrays = o_set_arr.shape[0]
    d_set_arr = np.zeros_like(o_set_arr)
    d_set_arr[:,0] = 0.
    d_set_arr[:,1] = 0.
    d_set_arr[:,2] = -1.
    o_set = gc.Point(o_set_arr)
    d_set = gc.Vector(d_set_arr)
    r_set = gc.Ray(o_set, d_set)

    p0 = gc.Point(msh.vertices[msh.faces[:,0],:])
    p1 = gc.Point(msh.vertices[msh.faces[:,1],:])
    p2 = gc.Point(msh.vertices[msh.faces[:,2],:])
    triangles = gc.Triangle(p0, p1, p2)
    ds_v2 = triangles.intersect_v2(r_set, diag_calc=True)
    ds_v3 = triangles.intersect_v3(r_set, diag_calc=True)

    t_v2, is_int_v2 = triangles.is_intersection_v2_t(r_set, diag_calc=True)
    t_v3, is_int_v3 = triangles.is_intersection_v3_t(r_set, diag_calc=True)
    assert (np.array_equal(ds_v2['is_intersection'].values, is_int_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['is_intersection'].values, is_int_v3, equal_nan=True))
    assert (np.array_equal(ds_v2['thit'].values, t_v2, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, t_v3, equal_nan=True))

    ndiag = nrays
    is_int_1d = np.full((ndiag), True, dtype=bool)
    t_1d = np.zeros((ndiag), dtype=np.float64)
    p_1d = np.zeros((ndiag,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for idiag in range (0, ndiag):
        p0 = gc.Point(msh.vertices[msh.faces[idiag,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[idiag,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[idiag,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        res_sca = triangle.intersect_v2(list_rays[idiag], ds_output=False)
        if (res_sca[2] is not None):
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[idiag] = ds_sca['is_intersection'].values
            t_1d[idiag] = ds_sca['thit'].values
            p_1d[idiag,:] = ds_sca['phit'].values
            n_1d[idiag,:] = ds_sca['nhit'].values
            dpdu_1d[idiag,:] = ds_sca['dpdu'].values
            dpdv_1d[idiag,:] = ds_sca['dpdv'].values
            u_1d[idiag] = ds_sca['u'].values
            v_1d[idiag] = ds_sca['v'].values
        else:
            is_int_1d[idiag] = False
            t_1d[idiag] = None
            p_1d[idiag,:] = None
            n_1d[idiag,:] = None
            dpdu_1d[idiag,:] = None
            dpdv_1d[idiag,:] = None
            u_1d[idiag] = None
            v_1d[idiag] = None

    assert (np.array_equal(ds_v2['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v2['dpdv'].values, dpdv_1d, equal_nan=True))

    ndiag = nrays
    is_int_1d = np.full((ndiag), True, dtype=bool)
    t_1d = np.zeros((ndiag), dtype=np.float64)
    p_1d = np.zeros((ndiag,3), dtype=np.float64)
    n_1d = np.zeros_like(p_1d)
    dpdu_1d = np.zeros_like(p_1d)
    dpdv_1d = np.zeros_like(p_1d)
    u_1d = np.zeros_like(t_1d)
    v_1d = np.zeros_like(t_1d)
    list_rays = []
    for ir in range (0, nrays):
        list_rays.append(gc.Ray(gc.Point(o_set_arr[ir,:]), gc.Vector(d_set_arr[ir,:])))

    for idiag in range (0, ndiag):
        p0 = gc.Point(msh.vertices[msh.faces[idiag,0],:])
        p1 = gc.Point(msh.vertices[msh.faces[idiag,1],:])
        p2 = gc.Point(msh.vertices[msh.faces[idiag,2],:])
        triangle = gc.Triangle(p0, p1, p2)
        res_sca = triangle.intersect_v3(list_rays[idiag], ds_output=False)
        if (res_sca[2] is not None):
            ds_sca = get_intersect_dataset(*res_sca)
            is_int_1d[idiag] = ds_sca['is_intersection'].values
            t_1d[idiag] = ds_sca['thit'].values
            p_1d[idiag,:] = ds_sca['phit'].values
            n_1d[idiag,:] = ds_sca['nhit'].values
            dpdu_1d[idiag,:] = ds_sca['dpdu'].values
            dpdv_1d[idiag,:] = ds_sca['dpdv'].values
            u_1d[idiag] = ds_sca['u'].values
            v_1d[idiag] = ds_sca['v'].values
        else:
            is_int_1d[idiag] = False
            t_1d[idiag] = None
            p_1d[idiag,:] = None
            n_1d[idiag,:] = None
            dpdu_1d[idiag,:] = None
            dpdv_1d[idiag,:] = None
            u_1d[idiag] = None
            v_1d[idiag] = None

    assert (np.array_equal(ds_v3['thit'].values, t_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['phit'].values, p_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['nhit'].values, n_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['u'].values, u_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['v'].values, v_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdu'].values, dpdu_1d, equal_nan=True))
    assert (np.array_equal(ds_v3['dpdv'].values, dpdv_1d, equal_nan=True))

    msh_ds = msh.intersect(r_set, diag_calc=True, use_loop=True)
    msh_dsn = msh.intersect(r_set, diag_calc=True, use_loop=False)
    assert (np.array_equal(msh_ds['thit'].values, msh_dsn['thit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['phit'].values, msh_dsn['phit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['nhit'].values, msh_dsn['nhit'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['u'].values, msh_dsn['u'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['v'].values, msh_dsn['v'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdu'].values, msh_dsn['dpdu'].values, equal_nan=True))
    assert (np.array_equal(msh_ds['dpdv'].values, msh_dsn['dpdv'].values, equal_nan=True))
    # Particular case where we should get the same values bellow (because diag calc)
    assert (np.array_equal(ds_v3['is_intersection'].values,
                           msh_ds['is_intersection'].values, equal_nan=True))
    assert (np.array_equal(ds_v3['thit'].values, msh_ds['thit'].values, equal_nan=True))

    thit, is_int = msh.is_intersection_t(r_set, diag_calc=True, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))
    assert (np.array_equal(msh_ds['thit'].values, thit, equal_nan=True))

    is_int = msh.is_intersection(r_set, diag_calc=True, use_loop=True)
    assert (np.array_equal(msh_ds['is_intersection'].values, is_int, equal_nan=True))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import geoclide as gc

PHIS = np.linspace(0., 360., 6)
THETAS = np.linspace(-180., 180., 7)


@pytest.mark.parametrize('phi', PHIS)
@pytest.mark.parametrize('theta', THETAS)
def test_vec2ang(phi, theta):
    v1 = gc.ang2vec(theta=theta, phi=phi, vec_view='zenith')
    th, ph = gc.vec2ang(v1, vec_view='zenith')
    v2 = gc.ang2vec(theta=th, phi=ph, vec_view='zenith')
    assert (np.isclose(v1.x, v2.x, 0., 1e-14))
    assert (np.isclose(v1.y, v2.y, 0., 1e-14))
    assert (np.isclose(v1.z, v2.z, 0., 1e-14))

    v1 = gc.ang2vec(theta=theta, phi=phi, vec_view='nadir')
    th, ph = gc.vec2ang(v1, vec_view='nadir')
    v2 = gc.ang2vec(theta=th, phi=ph, vec_view='nadir')
    assert (np.isclose(v1.x, v2.x, 0., 1e-14))
    assert (np.isclose(v1.y, v2.y, 0., 1e-14))
    assert (np.isclose(v1.z, v2.z, 0., 1e-14))

    # test standard values
    v1 = gc.ang2vec(theta=theta, phi=phi)
    th, ph = gc.vec2ang(v1)
    v2 = gc.ang2vec(theta=th, phi=ph)
    th, ph = gc.vec2ang(10*v1) # to test also non normalized vector
    v3 = gc.ang2vec(theta=th, phi=ph)
    assert (np.isclose(v1.x, v2.x, 0., 1e-14))
    assert (np.isclose(v1.y, v2.y, 0., 1e-14))
    assert (np.isclose(v1.z, v2.z, 0., 1e-14))
    assert (np.isclose(v1.x, v3.x, 0., 1e-14))
    assert (np.isclose(v1.y, v3.y, 0., 1e-14))
    assert (np.isclose(v1.z, v3.z, 0., 1e-14))

def test_vec2ang_old_bug():
    # check if bug from 3.0.1 is fixed
    gc.vec2ang(gc.Vector(-0.028830092518596997, 0.004805015419766166, -0.9995727775365759))


def test_ang2vec_diag():
    theta = np.array([20., 30., 45.])
    phi = np.array([0., 0., 180.])
    v_arr_f = gc.ang2vec(theta, phi, diag_calc=True).to_numpy()
    v_arr = np.zeros_like(v_arr_f)
    for i in range (0, len(theta)):
        v_arr[i,:] = gc.ang2vec(theta[i], phi[i]).to_numpy()
    assert (np.allclose(v_arr, v_arr_f))


def test_ang2vec_2d_arr():
    theta = np.array([20., 30., 45.])
    phi = np.array([0., 0., 180.])
    v_arr_f = gc.ang2vec(theta, phi).to_numpy()
    v_arr = np.zeros_like(v_arr_f)
    k = 0
    for j in range (0, len(phi)):
        for i in range (0, len(theta)):
            v_arr[k,:] = gc.ang2vec(theta[i], phi[j]).to_numpy()
            k+=1
    assert (np.allclose(v_arr, v_arr_f)) 


def test_ang2vec_1d_arr1():
    theta = np.array([20., 30., 45.])
    phi = 0.
    v_arr_f = gc.ang2vec(theta, phi).to_numpy()
    v_arr = np.zeros_like(v_arr_f)
    for i in range (0, len(theta)):
        v_arr[i,:] = gc.ang2vec(theta[i], phi).to_numpy()

    assert (np.allclose(v_arr, v_arr_f))


def test_ang2vec_1d_arr2():
    theta = 20.
    phi = np.array([0., 67., 180.])
    v_arr_f = gc.ang2vec(theta, phi).to_numpy()
    v_arr = np.zeros_like(v_arr_f)
    for j in range (0, len(phi)):
        v_arr[j,:] = gc.ang2vec(theta, phi[j]).to_numpy()

    assert (np.allclose(v_arr, v_arr_f))

def test_vec2ang_arr():
    phis = np.linspace(0., 360., 6)
    thetas = np.linspace(-180., 180., 6)
    v_set1 = gc.ang2vec(theta=thetas, phi=phis, vec_view='zenith', diag_calc=True)
    ths, phs = gc.vec2ang(v_set1, vec_view='zenith')
    v_set2 = gc.ang2vec(theta=ths, phi=phs, vec_view='zenith', diag_calc=True)

    assert (np.allclose(v_set1.to_numpy(), v_set2.to_numpy(), 0., 1e-15))

    v_set1 = gc.ang2vec(theta=thetas, phi=phis, vec_view='nadir', diag_calc=True)
    ths, phs = gc.vec2ang(v_set1, vec_view='nadir')
    v_set2 = gc.ang2vec(theta=ths, phi=phs, vec_view='nadir', diag_calc=True)

    assert (np.allclose(v_set1.to_numpy(), v_set2.to_numpy(), 0., 1e-15))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import math
import geoclide as gc


def test_dot():
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Vector(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Vector(0., 0., 1.)) == 1.)
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)
    v_set1 = gc.Vector(np.vstack((np.array([0., 0., 1.]), np.array([0., 0., 1.]))))
    v_set2 = gc.Vector(np.vstack((np.array([0., 1., 0.]), np.array([0., 0., 1.]))))
    assert (gc.dot(v_set1, v_set2)[0] == 0.)
    assert (gc.dot(v_set1, v_set2)[1] == 1.)

    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Vector(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Vector(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)

    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Normal(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Vector(0., 0., 1.), gc.Normal(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Vector(0., 0., 1.), v) == math.sqrt(2.)/2.)

    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Normal(0., 1., 0.)) == 0.)
    assert (gc.dot(gc.Normal(0., 0., 1.), gc.Normal(0., 0., 1.)) == 1.)
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.dot(gc.Normal(0., 0., 1.), v) == math.sqrt(2.)/2.)


def test_cross():
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Vector(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Vector(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Vector(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))
    v_set1 = gc.Vector(np.vstack((np.array([0., 0., 1.]), np.array([0., 0., 1.]))))
    v_set2 = gc.Vector(np.vstack((np.array([0., 1., 0.]), np.array([0., 0., 1.]))))
    assert (np.all(gc.cross(v_set1, v_set2).to_numpy()[0,:] == np.array([-1.0, 0.0, 0.0])))
    assert (np.all(gc.cross(v_set1, v_set2).to_numpy()[1,:] == np.array([0.0, 0.0, 0.0])))

    assert (gc.cross(gc.Normal(0., 0., 1.), gc.Vector(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Normal(0., 0., 1.), gc.Vector(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Vector(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Normal(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))

    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Normal(0., 1., 0.)) == gc.Vector(-1.0, 0.0, 0.0))
    assert (gc.cross(gc.Vector(0., 0., 1.), gc.Normal(0., 0., 1.)) == gc.Vector(0.0, 0.0, 0.0))
    v = gc.Normal(math.sqrt(2.)/2., 0., math.sqrt(2.)/2.)
    assert (gc.cross(gc.Vector(0., 0., 1.), v) ==  gc.Vector(0.0, math.sqrt(2.)/2., 0.0))


def test_normalize():
    length = math.sqrt(1. + 4. + 9.)
    a_1 = 1. / length
    a_2 = 2./ length
    a_3 = 3. / length
    assert (gc.normalize(gc.Vector(1.,2.,3.)) == gc.Vector(a_1, a_2, a_3))
    assert (gc.normalize(gc.Vector(-1.,-2.,-3.)) == gc.Vector(-a_1, -a_2, -a_3))
    v_set = gc.Vector(np.vstack((np.array([1.,2.,3.]), np.array([-1.,-2.,-3.]))))
    assert (np.all(gc.normalize(v_set).to_numpy()[0,:] == np.array([a_1, a_2, a_3])))
    assert (np.all(gc.normalize(v_set).to_numpy()[1,:] == np.array([-a_1, -a_2, -a_3])))

    assert (gc.normalize(gc.Normal(1.,2.,3.)) == gc.Normal(a_1, a_2, a_3))
    assert (gc.normalize(gc.Normal(-1.,-2.,-3.)) == gc.Normal(-a_1, -a_2, -a_3))


def test_coordinate_system():
    v1 = gc.Vector(0., 0., 1.)
    v2_m1, v3_m1 = gc.coordinate_system(v1, 'm1')
    assert (v2_m1 == gc.Vector(0.0, 1.0, -0.0))
    assert (v3_m1 == gc.Vector(-1.0, 0.0, 0.0))
    v2_m2, v3_m2 = gc.coordinate_system(v1, 'm2')
    assert (v2_m2 == gc.Vector(1.0, -0.0, -0.0))
    assert (v3_m2 == gc.Vector(-0.0, 1.0, -0.0))

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(4,2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    v1_u2, v1_u3 = gc.coordinate_system(v1, 'm1')
    v2_u2, v2_u3 = gc.coordinate_system(v2, 'm1')
    v1v2_u2, v1v2_u3 = gc.coordinate_system(v1v2, 'm1')
    assert(v1_u2 == gc.Vector(v1v2_u2.to_numpy()[0,:]))
    assert(v1_u3 == gc.Vector(v1v2_u3.to_numpy()[0,:]))
    assert(v2_u2 == gc.Vector(v1v2_u2.to_numpy()[1,:]))
    assert(v2_u3 == gc.Vector(v1v2_u3.to_numpy()[1,:]))


def test_distance():
    p1 = gc.Point(0., 0., 0.)
    p2 = gc.Point(0., 0., 10.)
    assert (gc.distance(p1,p2) == 10.)
    p1 = gc.Point(1., 2., 1.9)
    p2 = gc.Point(5., 15., 3.)
    a_1 = math.sqrt(16 + 13**2 + 1.1**2)
    assert (gc.distance(p1,p2) == a_1)
    p_set1 = gc.Point(np.array([[0., 0., 0.], [2., 3., 5.]]))
    p_set2 = gc.Point(np.array([[-5., -2., 0.], [-2., 3., 5.]]))
    p1 = gc.Point(p_set1.to_numpy()[0,:])
    p2 = gc.Point(p_set1.to_numpy()[1,:])
    p3 = gc.Point(p_set2.to_numpy()[0,:])
    p4 = gc.Point(p_set2.to_numpy()[1,:])
    assert (gc.distance(p_set1, p_set2)[0] == gc.distance(p1,p3))
    assert (gc.distance(p_set1, p_set2)[1] == gc.distance(p2,p4))


def test_face_forward():
    n1 = gc.Normal(1., 0., 0.)
    v1 = gc.Vector(-1., 0., 0.)
    assert (gc.face_forward(v1, n1) == gc.Vector(1.0, -0.0, -0.0))
    n1 = gc.Normal(-1., 0., 0.)
    v1 = gc.Vector(-1., 0., 0.)
    assert (gc.face_forward(v1, n1) == gc.Vector(-1.0, 0.0, 0.0))

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    v2v1 = gc.Vector(v1v2.to_numpy()[::-1,:])
    assert(gc.face_forward(v1, v2) == gc.Vector(gc.face_forward(v1v2, v2v1).to_numpy()[0,:]))
    assert(gc.face_forward(v2, v1) == gc.Vector(gc.face_forward(v1v2, v2v1).to_numpy()[1,:]))


def test_vmax():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vmax(v1) == 3.)

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert(gc.vmax(v1) == gc.vmax(v1v2)[0])
    assert(gc.vmax(v2) == gc.vmax(v1v2)[1])


def test_vmin():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vmin(v1) == 1.)

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert(gc.vmin(v1) == gc.vmin(v1v2)[0])
    assert(gc.vmin(v2) == gc.vmin(v1v2)[1])


def test_vargmax():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vargmax(v1) == 1)

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert(gc.vargmax(v1) == gc.vargmax(v1v2)[0])
    assert(gc.vargmax(v2) == gc.vargmax(v1v2)[1])


def test_vargmin():
    v1 = gc.Vector(2.,3.,1.)
    assert (gc.vargmin(v1) == 2)

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert(gc.vargmin(v1) == gc.vargmin(v1v2)[0])
    assert(gc.vargmin(v2) == gc.vargmin(v1v2)[1])


def test_vabs():
    v1 = gc.Vector(5,-2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert (gc.vabs(v1) == gc.Vector(5, 2, 10))
    assert (gc.vabs(v2) == gc.Vector(4, 2, 1))
    assert (gc.Vector(gc.vabs(v1v2).to_numpy()[0,:]) == gc.Vector(5, 2, 10))
    assert (gc.Vector(gc.vabs(v1v2).to_numpy()[1,:]) == gc.Vector(4, 2, 1))

def test_permute():
    v1 = gc.Vector(2., 3., 1.)
    assert (gc.permute(v1, 1, 0, 2) == gc.Vector(3.0, 2.0, 1.0))

    v1 = gc.Vector(5,2,10)
    v2 = gc.Vector(-4,-2,-1)
    v1v2 = gc.Vector(np.vstack((v1.to_numpy(), v2.to_numpy())))
    assert(gc.permute(v1, 1, 2, 0) == gc.Vector(gc.permute(v1v2, 1, 2 , 0).to_numpy()[0,:]))
    assert(gc.permute(v2, 1, 2, 0) == gc.Vector(gc.permute(v1v2, 1, 2 , 0).to_numpy()[1,:]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import math


def clamp(val, val_min, val_max):
    """
    Clamps val into the range [val_min, val_max]

    Parameters
    ----------
    val : float
        The scalar to be clamped
    val_min : float
        The minumum value
    val_max : float
        The maximum value

    Returns
    -------
    out : float
        The result of the clamp

    Examples
    --------
    >>> import geoclide as gc
    >>> gc.clamp(4, val_min=5, val_max=11)
    5
    """
    if (not np.isscalar(val)     or 
        not np.isscalar(val_min) or
        not np.isscalar(val_max) ):
        raise ValueError('The parameters must be all scalars')
    
    return val_min if (val < val_min) else (val_max if (val > val_max) else val)


def quadratic(a, b, c):
    """
    Resolve the quadratic polynomial: ax**2 + bx + c

    - where x is the quadratic polynomial variable and a, b and c the coefficients

    Parameters
    ----------
    a : float | 1-D ndarray
        The first coefficient(s) of the quadratic polynomial
    b : float | 1-D ndarray
        The second coefficient(s) of the quadratic polynomial
    c : float | 1-D ndarray
        The third coefficient(s) of the quadratic polynomial

    Returns
    -------
    b : bool | 1-D ndarray
        If the quadratic can be solved return True, else False
    x0 : float | None | 1-D ndarray
        The first solution(s)
    x1 : float | None | 1-D ndarray
        The second solution(s)

    Notes
    -----
    If There are 2 solutions x0 < x1. And if there is only one solution x0 = x1.

    Examples
    --------
    >>> import geoclide as gc
    >>> a = 2
    >>> b = -5
    >>> c = 0
    >>> gc.quadratic(a, b, c)
    (True, 0.0, 2.5)
    """
    if isinstance(a, np.ndarray):        
        #  Find quadratic discriminant
        discrim = (b * b) - (4 * a * c)
        is_solution = np.full(discrim.shape, True, dtype=bool)

        c1 = discrim < 0
        rootDiscrim = np.sqrt(discrim)

        # Compute quadratic xi values
        q = np.zeros_like(discrim)

        c2 = b < 0
        not_c2 = np.logical_not(c2)
        q[c2] = -0.5 * (b[c2] - rootDiscrim[c2])
        q[not_c2] = -0.5 * (b[not_c2] + rootDiscrim[not_c2])

        x0 = np.zeros_like(discrim)
        c3 = a!=0
        not_c3 = np.logical_not(c3)
        x0[c3] =  q[c3] / a[c3]
        x0[not_c3] = c[not_c3] / q[not_c3]

        x1 = c / q

        c4 = x0 > x1
        x0[c4], x1[c4] = x1[c4], x0[c4]

        is_solution[c1] = False
        x0[c1] = None
        x1[c1] = None

        return is_solution, x0, x1
    else:
        #  Find quadratic discriminant
        discrim = (b * b) - (4 * a * c)

        if (discrim < 0): return False, None, None

        rootDiscrim = math.sqrt(discrim)

        # Compute quadratic xi values
        if (b < 0): q = -0.5 * (b - rootDiscrim)
        else: q = -0.5 * (b + rootDiscrim)

        if (a != 0): x0 = q / a
        else: x0 = c / q

        x1 = c / q

        if (x0 > x1): x0, x1 = x1, x0

        return True, x0, x1


def gamma_f32(n):
    """
    :meta private:

    Gamma function from pbrt v3
    """
    epsi = np.finfo(np.float32).eps * 0.5
    return (n*epsi)/(1 - n*epsi)


def gamma_f64(n):
    """
    :meta private:

    Gamma function from pbrt v3 but in double precision
    """
    epsi = np.finfo(np.float64).eps * 0.5
    return (n*epsi)/(1 - n*epsi)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from geoclide.shapes import Shape, get_intersect_dataset
from geoclide.mathope import clamp, quadratic
from geoclide.vecope import distance
from geoclide.basic import Ray, Vector, Point
from geoclide.transform import Transform, get_scale_tf
from geoclide.trianglemesh import create_sphere_trianglemesh, TriangleMesh, create_disk_trianglemesh
import math
import numpy as np
from geoclide.constante import TWO_PI


class Sphere(Shape):
    '''
    Creation of the class Sphere

    - without transformation the sphere is centered at the origin
    - z0, z1 and phi_max are needed parameters for the creation of any partial sphere

    Parameters
    ----------
    radius : float
        The radius of the sphere
    z_min : float, optional
        The minimum z value of the sphere where z0 is between [-radius, 0]
    z_max : float, optional
        The maximum z value of the sphere where z1 is between [0, radius]
    phi_max : float, optional
        The maximum phi value in degrees of the sphere, where phi is between 0 and 360°
    oTw : Transform, optional
        From object to world space or the transformation applied to the sphere
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the sphere
    '''
    def __init__(self, radius, z_min=None, z_max=None, phi_max=360., oTw=None, wTo=None):
        if z_min is None: z_min = -radius
        if z_max is None: z_max = radius
        if z_max < z_min : raise ValueError ('zmax must be greater than zmin')
        if (phi_max < 0. or phi_max > 360.):
            raise ValueError ('The value of the parameter phi_max must in the range: [0, 360]')
        if wTo is None and oTw is None:
            wTo = Transform()
            oTw = Transform()
        elif (wTo is None and isinstance(oTw, Transform)): wTo = oTw.inverse()
        elif (isinstance(wTo, Transform) and oTw is None): oTw = wTo.inverse()
        if (not np.isscalar(radius) or
            not np.isscalar(z_min)  or
            not np.isscalar(z_max)  or
            not np.isscalar(phi_max) ):
            raise ValueError('The parameters radius, z_min, z_max and phi_max must be all scalars')
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.radius = radius
        self.zmin = clamp(z_min, -self.radius, self.radius)
        self.zmax = clamp(z_max, -self.radius, self.radius)
        self.theta_min = math.acos(clamp(self.zmin/self.radius, -1, 1))
        self.theta_max = math.acos(clamp(self.zmax/self.radius, -1, 1))
        self.phi_max = phi_max

    def is_intersection_t(self, r):
        """
        Test if a ray/set of rays intersects the sphere/partial sphere

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        thit : float | 1-D ndarray
            The t ray variable(s) for its first intersection at the shape surface
        is_intersection : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> sph1 = gc.Sphere(radius=1.) # sphere of radius 1
        >>> sph2 = gc.Sphere(radius=1., z_max=0.5) # partial sphere where portion above z=0.5 is removed
        >>> r = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
        >>> sph1.is_intersection_t(r)
        (1.4000000000000004, True)
        >>> sph2.is_intersection(r) # here no intersection since the sphere part above z=0.5 is removed
        False
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # Compute quadratic sphere coefficients
                a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
                b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
                c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
                    self.radius*self.radius

                # Solve quadratic equation
                exist, t0, t1 = quadratic(a, b, c)
                c1 = np.logical_not(exist)
                
                # Compute intersection distance along ray
                c2 = np.logical_or(t0 > ray.maxt, t1 < ray.mint)
                thit = t0

                c3_bis_1 = t0 < ray.mint
                thit[c3_bis_1] = t1[c3_bis_1]
                c3_bis_2 = thit > ray.maxt
                c3 = np.logical_and(c3_bis_1, c3_bis_2)

                # Compute sphere hit position and $\phi$
                phit = ray(thit)
                phit *= self.radius / distance(phit, Point(0., 0., 0.))
                phit.x[np.logical_and(phit.x == 0, phit.y == 0)] = 1e-5 * self.radius
                phi = np.arctan2(phit.y, phit.x)
                phi[phi < 0] += TWO_PI

                # Test sphere intersection against clipping parameters
                phi_max_rad = math.radians(self.phi_max)
                c4_bis_1 = np.logical_and(self.zmin > -self.radius, phit.z < self.zmin)
                c4_bis_2 = np.logical_and(self.zmax <  self.radius, phit.z > self.zmax)
                c4_bis_3 = phi > phi_max_rad
                c4 = np.logical_or.reduce((c4_bis_1, c4_bis_2, c4_bis_3))
                c5 = np.logical_and(c4, np.logical_or(thit == t1, t1 > ray.maxt))

                if np.any(c4):
                    thit[c4] = t1[c4]
                    # Compute sphere hit position and $\phi$
                    phit_bis = ray[thit]
                    phit.x[c4] = phit_bis.x[c4]
                    phit.y[c4] = phit_bis.y[c4]
                    phit.z[c4] = phit_bis.z[c4]
                    c4_p1 = np.logical_and.reduce((c4, phit.x == 0, phit.y == 0))
                    phit.x[c4_p1] = 1e-5 * self.radius
                    phi[c4] = np.arctan2(phit.y[c4], phit.x[c4])
                    c4_p2 = np.logical_and(c4, phi < 0)
                    phi[c4_p2] += TWO_PI
                
                c6_bis_1 = np.logical_and(self.zmin > -self.radius, phit.z < self.zmin)
                c6_bis_2 = np.logical_and(self.zmax <  self.radius, phit.z > self.zmax)
                c6_bis_3 = phi > phi_max_rad
                c6 = np.logical_and(c4, np.logical_or.reduce((c6_bis_1, c6_bis_2, c6_bis_3)))

                c7 = np.logical_or.reduce((c1, c2, c3, c5, c6))
                is_intersection[c7] = False
                thit[c7] = None

                return thit, is_intersection
        else:
            # Compute quadratic sphere coefficients
            a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
            b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
            c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
                self.radius*self.radius

            # Solve quadratic equation
            exist, t0, t1 = quadratic(a, b, c)
            if (not exist): return None, False
            
            # Compute intersection distance along ray
            if (t0 > ray.maxt or t1 < ray.mint): return None, False
            thit = t0

            if (t0 < ray.mint):
                thit = t1
                if (thit > ray.maxt): return None, False

            # Compute sphere hit position and $\phi$
            phit = ray(thit)
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
            phi = math.atan2(phit.y, phit.x)
            if (phi < 0): phi += TWO_PI

            # Test sphere intersection against clipping parameters
            phi_max_rad = math.radians(self.phi_max)
            if ((self.zmin > -self.radius and phit.z < self.zmin) or
                (self.zmax <  self.radius and phit.z > self.zmax) or
                (phi > phi_max_rad) ):
                if (thit == t1): return None, False
                if (t1 > ray.maxt): return None ,False
                thit = t1
                # Compute sphere hit position and $\phi$
                phit = ray(thit)
                if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
                phi = math.atan2(phit.y, phit.x)
                if (phi < 0): phi += TWO_PI
                if ((self.zmin > -self.radius and phit.z < self.zmin) or
                    (self.zmax <  self.radius and phit.z > self.zmax) or
                    (phi > phi_max_rad) ):
                    return None, False

            return thit, True
    
    def is_intersection(self, r):
        """
        Test if a ray/set of rays intersects the sphere / partial sphere

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        out : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> sph1 = gc.Sphere(radius=1.) # sphere of radius 1
        >>> sph2 = gc.Sphere(radius=1., z_max=0.5) # partial sphere where portion above z=0.5 is removed
        >>> r = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
        >>> sph1.is_intersection(r)
        True
        >>> sph2.is_intersection(r) # here no intersection since the sphere part above z=0.5 is removed
        False
        """
        _, is_intersection = self.is_intersection_t(r)
        return is_intersection

    def intersect(self, r, ds_output=True):
        """
        Test if a ray/set of rays intersects the sphere/partial sphere

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test(s)
        ds_output : Bool, optional
            If True the output is a dataset, else -> a tuple with intersection information variables

        Returns
        -------
        out : xr.Dataset | tuple
            Look-up table with the intersection information if ds_output is True, 
            else returns a tuple. The tuple is ready to be an input for the function 
            geoclide.shapes.get_intersect_dataset

        Examples
        --------
        >>> import geoclide as gc
        >>> sph1 = gc.Sphere(radius=1.) # sphere of radius 1
        >>> sph2 = gc.Sphere(radius=1., z_max=0.5) # partial sphere where portion above z=0.5 is removed
        >>> r = gc.Ray(o=gc.Point(-2., 0., 0.8), d=gc.Vector(1.,0.,0.))
        >>> sph1.intersect(r)
        <xarray.Dataset> Size: 209B
        Dimensions:          (xyz: 3)
        Coordinates:
        * xyz              (xyz) int64 24B 0 1 2
        Data variables:
            o                (xyz) float64 24B -2.0 0.0 0.8
            d                (xyz) float64 24B 1.0 0.0 0.0
            mint             int64 8B 0
            maxt             float64 8B inf
            is_intersection  bool 1B True
            thit             float64 8B 1.4
            u                float64 8B 0.5
            v                float64 8B 0.7952
            phit             (xyz) float64 24B -0.6 0.0 0.8
            nhit             (xyz) float64 24B -0.6 0.0 0.8
            dpdu             (xyz) float64 24B 0.0 -3.77 0.0
            dpdv             (xyz) float64 24B 2.513 0.0 1.885
        >>> ds = sph1.intersect(r)
        >>> ds['phit'].values # the intersection point
        array([-0.6,  0. ,  0.8])
        >>> ds['nhit'].values # The surface normal at the intersection point
        array([-0.6,  0. ,  0.8])
        >>> sph2.intersect(r) # here no intersection since the sphere part above z=0.5 is removed
        <xarray.Dataset> Size: 209B
        Dimensions:          (xyz: 3)
        Coordinates:
        * xyz              (xyz) int64 24B 0 1 2
        Data variables:
            o                (xyz) float64 24B -2.0 0.0 0.8
            d                (xyz) float64 24B 1.0 0.0 0.0
            mint             int64 8B 0
            maxt             float64 8B inf
            is_intersection  bool 1B False
            thit             object 8B None
            u                object 8B None
            v                object 8B None
            phit             (xyz) float64 24B nan nan nan
            nhit             (xyz) float64 24B nan nan nan
            dpdu             (xyz) float64 24B nan nan nan
            dpdv             (xyz) float64 24B nan nan nan
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        sh_name = self.__class__.__name__
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # Compute quadratic sphere coefficients
                a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
                b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
                c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
                    self.radius*self.radius

                # Solve quadratic equation
                exist, t0, t1 = quadratic(a, b, c)
                c1 = np.logical_not(exist)
                
                # Compute intersection distance along ray
                c2 = np.logical_or(t0 > ray.maxt, t1 < ray.mint)
                thit = t0

                c3_bis_1 = t0 < ray.mint
                thit[c3_bis_1] = t1[c3_bis_1]
                c3_bis_2 = thit > ray.maxt
                c3 = np.logical_and(c3_bis_1, c3_bis_2)

                # Compute sphere hit position and $\phi$
                phit = ray(thit)
                phit *= self.radius / distance(phit, Point(0., 0., 0.))
                phit.x[np.logical_and(phit.x == 0, phit.y == 0)] = 1e-5 * self.radius
                phi = np.arctan2(phit.y, phit.x)
                phi[phi < 0] += TWO_PI

                # Test sphere intersection against clipping parameters
                phi_max_rad = math.radians(self.phi_max)
                c4_bis_1 = np.logical_and(self.zmin > -self.radius, phit.z < self.zmin)
                c4_bis_2 = np.logical_and(self.zmax <  self.radius, phit.z > self.zmax)
                c4_bis_3 = phi > phi_max_rad
                c4 = np.logical_or.reduce((c4_bis_1, c4_bis_2, c4_bis_3))
                c5 = np.logical_and(c4, np.logical_or(thit == t1, t1 > ray.maxt))

                if np.any(c4):
                    thit[c4] = t1[c4]
                    # Compute sphere hit position and $\phi$
                    phit_bis = ray[thit]
                    phit.x[c4] = phit_bis.x[c4]
                    phit.y[c4] = phit_bis.y[c4]
                    phit.z[c4] = phit_bis.z[c4]
                    c4_p1 = np.logical_and.reduce((c4, phit.x == 0, phit.y == 0))
                    phit.x[c4_p1] = 1e-5 * self.radius
                    phi[c4] = np.arctan2(phit.y[c4], phit.x[c4])
                    c4_p2 = np.logical_and(c4, phi < 0)
                    phi[c4_p2] += TWO_PI
                
                c6_bis_1 = np.logical_and(self.zmin > -self.radius, phit.z < self.zmin)
                c6_bis_2 = np.logical_and(self.zmax <  self.radius, phit.z > self.zmax)
                c6_bis_3 = phi > phi_max_rad
                c6 = np.logical_and(c4, np.logical_or.reduce((c6_bis_1, c6_bis_2, c6_bis_3)))

                # Find parametric representation of sphere hit
                u = phi / phi_max_rad
                theta = np.arccos(np.clip(phit.z / self.radius, -1, 1))
                v = (theta - self.theta_min) / (self.theta_max - self.theta_min)

                # Compute sphere $\dpdu$ and $\dpdv$
                zradius = np.sqrt(phit.x*phit.x + phit.y*phit.y)
                invzradius = 1 / zradius
                cosphi = phit.x * invzradius
                sinphi = phit.y * invzradius
                zeros = np.zeros(nrays, dtype=np.float64)
                dpdu = Vector(-phi_max_rad * phit.y, phi_max_rad * phit.x, zeros)
                dpdv = (self.theta_max-self.theta_min) * Vector(phit.z*cosphi, phit.z*sinphi, -self.radius*np.sin(theta))

                c7 = np.logical_or.reduce((c1, c2, c3, c5, c6))
                is_intersection[c7] = False
                thit[c7] = None

                out = sh_name, r, thit, is_intersection, u, v, self.oTw(dpdu).to_numpy(), self.oTw(dpdv).to_numpy(), False
                if ds_output : return get_intersect_dataset(*out)
                else : return out
        else:
            # Compute quadratic sphere coefficients
            a = ray.d.x*ray.d.x + ray.d.y*ray.d.y + ray.d.z*ray.d.z
            b = 2 * (ray.d.x*ray.o.x + ray.d.y*ray.o.y + ray.d.z*ray.o.z)
            c = ray.o.x*ray.o.x + ray.o.y*ray.o.y + ray.o.z*ray.o.z - \
                self.radius*self.radius

            # Solve quadratic equation
            exist, t0, t1 = quadratic(a, b, c)
            if (not exist):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            
            # Compute intersection distance along ray
            if (t0 > ray.maxt or t1 < ray.mint):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            thit = t0

            if (t0 < ray.mint):
                thit = t1
                if (thit > ray.maxt):
                    if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                    else : return sh_name, r, None, False, None, None, None, None, False

            # Compute sphere hit position and $\phi$
            phit = ray(thit)
            phit *= self.radius / distance(phit, Point(0., 0., 0.))
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
            phi = math.atan2(phit.y, phit.x)
            if (phi < 0): phi += TWO_PI

            # Test sphere intersection against clipping parameters
            phi_max_rad = math.radians(self.phi_max)
            if ((self.zmin > -self.radius and phit.z < self.zmin) or
                (self.zmax <  self.radius and phit.z > self.zmax) or
                (phi > phi_max_rad) ):
                if (thit == t1 or t1 > ray.maxt):
                    if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                    else : return sh_name, r, None, False, None, None, None, None, False
                thit = t1
                # Compute sphere hit position and $\phi$
                phit = ray(thit)
                if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.radius
                phi = math.atan2(phit.y, phit.x)
                if (phi < 0): phi += TWO_PI
                if ((self.zmin > -self.radius and phit.z < self.zmin) or
                    (self.zmax <  self.radius and phit.z > self.zmax) or
                    (phi > phi_max_rad) ):
                    if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                    else : return sh_name, r, None, False, None, None, None, None, False

            # Find parametric representation of sphere hit
            u = phi / phi_max_rad
            theta = math.acos(clamp(phit.z / self.radius, -1, 1))
            v = (theta - self.theta_min) / (self.theta_max - self.theta_min)

            # Compute sphere $\dpdu$ and $\dpdv$
            zradius = math.sqrt(phit.x*phit.x + phit.y*phit.y)
            invzradius = 1 / zradius
            cosphi = phit.x * invzradius
            sinphi = phit.y * invzradius
            dpdu = Vector(-phi_max_rad * phit.y, phi_max_rad * phit.x, 0)
            dpdv = (self.theta_max-self.theta_min) * Vector(phit.z*cosphi, phit.z*sinphi, -self.radius*math.sin(theta)) 

            out = sh_name, r, thit, True, u, v, self.oTw(dpdu).to_numpy(), self.oTw(dpdv).to_numpy(), False
            if ds_output : return get_intersect_dataset(*out)
            else : return out

    def area(self):
        """
        compute the sphere / partial sphere area

        .. warning::
            `the scale transformation is not considered for the area calculation!`
        """
        return (math.radians(self.phi_max) * self.radius * (self.zmax-self.zmin)) # The sphere / partial sphere area
    
    def to_trianglemesh(self, reso_theta=None, reso_phi=None):
        """
        Convert the sphere to a triangle mesh

        Parameters
        ----------
        reso_theta : int, optional
            The number of lines around the polar theta angle, minimum accepted value is 3
        reso_phi : int, optional
            The number of lines around the azimuth phi angle, minimum accepted value is 3

        Returns
        -------
        mesh : TriangleMesh
            The sphere converted to a triangle mesh
        """
        theta_zmin = clamp(math.degrees(self.theta_min), 0., 360.)
        theta_zmax = clamp(math.degrees(self.theta_max), 0., 360.)
        theta_min = min(theta_zmin, theta_zmax)
        theta_max = max(theta_zmin, theta_zmax)
        return create_sphere_trianglemesh(self.radius, reso_theta, reso_phi, theta_min, theta_max,
                                          self.phi_max, self.oTw, self.wTo)
    
    def plot(self, **kwargs):
        """
        Plot the sphere

        - The sphere is first converted to a triangle mesh then the TriangleMesh 
          plot method is used

        Parameters
        ----------
        **kwargs
            The keyword arguments are passed on to the TriangleMesh plot method
        """
        return self.to_trianglemesh().plot(**kwargs)


class Spheroid(Shape):
    '''
    Creation of the class Spheroid

    - without transformation the spheroid is centered at the origin
    - spheroid equation: x/(alpha**2) + y/(alpha**2) + z/(gamma**2) = 1,
      where alpha = radius_xy and gamma = radius_z
    - prolate -> radius_z > radius_xy
    - oblate -> radius_z < radius_xy

    Parameters
    ----------
    radius_xy : float
        The equatorial radius of the spheroid
    radius_z : float
        The pole radius of the spheroid (distance from center to pole along z axis)
    oTw : Transform, optional
        From object to world space or the transformation applied to the spheroid
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the spheroid
    '''
    def __init__(self, radius_xy, radius_z, oTw=None, wTo=None):
        if wTo is None and oTw is None:
            wTo = Transform()
            oTw = Transform()
        elif (wTo is None and isinstance(oTw, Transform)): wTo = oTw.inverse()
        elif (isinstance(wTo, Transform) and oTw is None): oTw = wTo.inverse()
        if (not np.isscalar(radius_xy) or not np.isscalar(radius_z)):
            raise ValueError('The parameters alpha and gamma must be all scalars')
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.alpha = radius_xy
        self.gamma = radius_z
        self.alpha2 = radius_xy*radius_xy
        self.gamma2 = radius_z*radius_z

    def is_intersection_t(self, r):
        """
        Test if a ray/set of rays intersects the spheroid

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        thit : float | 1-D ndarray
            The t ray variable(s) for its first intersection at the shape surface
        is_intersection : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> oblate = gc.Spheroid(radius_xy=3., radius_z=1.5)
        >>> prolate = gc.Spheroid(radius_xy=1.5, radius_z=3.)
        >>> r1 = gc.Ray(o=gc.Point(2.5, 0., 10.), d=(gc.Vector(0., 0., -1.)))
        >>> r2 = gc.Ray(o=gc.Point(10., 0., 2.5), d=(gc.Vector(-1., 0., 0.)))
        >>> oblate.is_intersection_t(r1)
        (9.170843802411135, True)
        >>> oblate.is_intersection_t(r2)
        (None, False)
        >>> prolate.is_intersection_t(r1)
        (None, False)
        >>> prolate.is_intersection_t(r2)
        (9.170843802411135, True)
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # Compute quadratic sphere coefficients
                inv_alpha2 = 1./self.alpha2
                inv_beta2 = inv_alpha2 # ellipsoid special case where alpha=beta
                inv_gamma2 = 1./self.gamma2
                a = ray.d.x*ray.d.x*inv_alpha2 + ray.d.y*ray.d.y*inv_beta2 + ray.d.z*ray.d.z*inv_gamma2
                b = 2 * (ray.d.x*ray.o.x*inv_alpha2 + ray.d.y*ray.o.y*inv_beta2 + ray.d.z*ray.o.z*inv_gamma2)
                c = ray.o.x*ray.o.x*inv_alpha2 + ray.o.y*ray.o.y*inv_beta2 + ray.o.z*ray.o.z*inv_gamma2 - 1

                # Solve quadratic equation
                exist, t0, t1 = quadratic(a, b, c)
                c1 = np.logical_not(exist)
                
                # Compute intersection distance along ray
                c2 = np.logical_or(t0 > ray.maxt, t1 < ray.mint)
                thit = t0

                c3_bis_1 = t0 < ray.mint
                thit[c3_bis_1] = t1[c3_bis_1]
                c3_bis_2 = thit > ray.maxt
                c3 = np.logical_and(c3_bis_1, c3_bis_2)

                c4 = np.logical_or.reduce((c1, c2, c3))
                is_intersection[c4] = False
                thit[c4] = None

                return thit, is_intersection
        else:
            # Compute quadratic sphere coefficients
            inv_alpha2 = 1./self.alpha2
            inv_beta2 = inv_alpha2 # ellipsoid special case where alpha=beta
            inv_gamma2 = 1./self.gamma2
            a = ray.d.x*ray.d.x*inv_alpha2 + ray.d.y*ray.d.y*inv_beta2 + ray.d.z*ray.d.z*inv_gamma2
            b = 2 * (ray.d.x*ray.o.x*inv_alpha2 + ray.d.y*ray.o.y*inv_beta2 + ray.d.z*ray.o.z*inv_gamma2)
            c = ray.o.x*ray.o.x*inv_alpha2 + ray.o.y*ray.o.y*inv_beta2 + ray.o.z*ray.o.z*inv_gamma2 - 1

            # Solve quadratic equation
            exist, t0, t1 = quadratic(a, b, c)
            if (not exist): return None, False
            
            # Compute intersection distance along ray
            if (t0 > ray.maxt or t1 < ray.mint): return None, False
            thit = t0

            if (t0 < ray.mint):
                thit = t1
                if (thit > ray.maxt): return None, False

            return thit, True
    
    def is_intersection(self, r):
        """
        Test if a ray/set of rays intersects the spheroid

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        out : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> oblate = gc.Spheroid(radius_xy=3., radius_z=1.5)
        >>> prolate = gc.Spheroid(radius_xy=1.5, radius_z=3.)
        >>> r1 = gc.Ray(o=gc.Point(2.5, 0., 10.), d=(gc.Vector(0., 0., -1.)))
        >>> r2 = gc.Ray(o=gc.Point(10., 0., 2.5), d=(gc.Vector(-1., 0., 0.)))
        >>> oblate.is_intersection(r1)
        True
        >>> oblate.is_intersection(r2)
        False
        >>> prolate.is_intersection(r1)
        False
        >>> prolate.is_intersection(r2)
        True
        """
        _, is_intersection = self.is_intersection_t(r)
        return is_intersection

    def intersect(self, r, ds_output=True):
        """
        Test if a ray/set of rays intersects the spheroid

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test(s)
        ds_output : Bool, optional
            If True the output is a dataset, else -> a tuple with intersection information variables

        Returns
        -------
        out : xr.Dataset | tuple
            Look-up table with the intersection information if ds_output is True, 
            else returns a tuple. The tuple is ready to be an input for the function 
            geoclide.shapes.get_intersect_dataset

        Examples
        --------
        >>> import geoclide as gc
        >>> oblate = gc.Spheroid(radius_xy=3., radius_z=1.5)
        >>> prolate = gc.Spheroid(radius_xy=1.5, radius_z=3.)
        >>> r1 = gc.Ray(o=gc.Point(2.5, 0., 10.), d=(gc.Vector(0., 0., -1.)))
        >>> r2 = gc.Ray(o=gc.Point(10., 0., 2.5), d=(gc.Vector(-1., 0., 0.)))
        >>> oblate.intersect(r1)
        <xarray.Dataset> Size: 209B
        Dimensions:          (xyz: 3)
        Coordinates:
        * xyz              (xyz) int64 24B 0 1 2
        Data variables:
            o                (xyz) float64 24B 2.5 0.0 10.0
            d                (xyz) float64 24B 0.0 0.0 -1.0
            mint             int64 8B 0
            maxt             float64 8B inf
            is_intersection  bool 1B True
            thit             float64 8B 9.171
            u                float64 8B 0.0
            v                float64 8B 0.6864
            phit             (xyz) float64 24B 2.5 0.0 0.8292
            nhit             (xyz) float64 24B 0.6019 -0.0 0.7985
            dpdu             (xyz) float64 24B 0.0 15.71 0.0
            dpdv             (xyz) float64 24B -5.21 0.0 3.927
        >>> prolate.intersect(r2)
        <xarray.Dataset> Size: 209B
        Dimensions:          (xyz: 3)
        Coordinates:
        * xyz              (xyz) int64 24B 0 1 2
        Data variables:
            o                (xyz) float64 24B 10.0 0.0 2.5
            d                (xyz) float64 24B -1.0 0.0 0.0
            mint             int64 8B 0
            maxt             float64 8B inf
            is_intersection  bool 1B True
            thit             float64 8B 9.171
            u                float64 8B 0.0
            v                float64 8B 0.8136
            phit             (xyz) float64 24B 0.8292 0.0 2.5
            nhit             (xyz) float64 24B 0.7985 -0.0 0.6019
            dpdu             (xyz) float64 24B 0.0 5.21 0.0
            dpdv             (xyz) float64 24B -3.927 0.0 5.21
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        sh_name = self.__class__.__name__
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # Compute quadratic sphere coefficients
                inv_alpha2 = 1./self.alpha2
                inv_beta2 = inv_alpha2 # ellipsoid special case where alpha=beta
                inv_gamma2 = 1./self.gamma2
                a = ray.d.x*ray.d.x*inv_alpha2 + ray.d.y*ray.d.y*inv_beta2 + ray.d.z*ray.d.z*inv_gamma2
                b = 2 * (ray.d.x*ray.o.x*inv_alpha2 + ray.d.y*ray.o.y*inv_beta2 + ray.d.z*ray.o.z*inv_gamma2)
                c = ray.o.x*ray.o.x*inv_alpha2 + ray.o.y*ray.o.y*inv_beta2 + ray.o.z*ray.o.z*inv_gamma2 - 1

                # Solve quadratic equation
                exist, t0, t1 = quadratic(a, b, c)
                c1 = np.logical_not(exist)
                
                # Compute intersection distance along ray
                c2 = np.logical_or(t0 > ray.maxt, t1 < ray.mint)
                thit = t0

                c3_bis_1 = t0 < ray.mint
                thit[c3_bis_1] = t1[c3_bis_1]
                c3_bis_2 = thit > ray.maxt
                c3 = np.logical_and(c3_bis_1, c3_bis_2)

                # Compute sphere hit position and $\phi$
                phit = ray(thit)
                phit.x[np.logical_and(phit.x == 0, phit.y == 0)] = 1e-5 * self.alpha
                phi = np.arctan2(phit.y, phit.x)
                phi[phi < 0] += TWO_PI

                # Find parametric representation of sphere hit
                u = phi / TWO_PI
                theta = np.arccos(np.clip(phit.z / self.gamma, -1, 1))
                v = 1 - (theta / math.pi)

                # Compute sphere $\dpdu$ and $\dpdv$
                zradius = np.sqrt(phit.x*phit.x + phit.y*phit.y)
                invzradius = 1 / zradius
                cosphi = phit.x * invzradius
                sinphi = phit.y * invzradius
                fac = -math.pi*(self.alpha/self.gamma)*phit.z
                zeros = np.zeros(nrays, dtype=np.float64)
                dpdu = Vector(-TWO_PI*phit.y, TWO_PI*phit.x, zeros)
                dpdv = Vector(fac*cosphi, fac*sinphi, math.pi*self.gamma*np.sin(theta))

                c4 = np.logical_or.reduce((c1, c2, c3))
                is_intersection[c4] = False
                thit[c4] = None

                out = sh_name, r, thit, is_intersection, u, v, self.oTw(dpdu).to_numpy(), self.oTw(dpdv).to_numpy(), False
                if ds_output : return get_intersect_dataset(*out)
                else : return out
        else:
            # Compute quadratic sphere coefficients
            inv_alpha2 = 1./self.alpha2
            inv_beta2 = inv_alpha2 # ellipsoid special case where alpha=beta
            inv_gamma2 = 1./self.gamma2
            a = ray.d.x*ray.d.x*inv_alpha2 + ray.d.y*ray.d.y*inv_beta2 + ray.d.z*ray.d.z*inv_gamma2
            b = 2 * (ray.d.x*ray.o.x*inv_alpha2 + ray.d.y*ray.o.y*inv_beta2 + ray.d.z*ray.o.z*inv_gamma2)
            c = ray.o.x*ray.o.x*inv_alpha2 + ray.o.y*ray.o.y*inv_beta2 + ray.o.z*ray.o.z*inv_gamma2 - 1 

            # Solve quadratic equation
            exist, t0, t1 = quadratic(a, b, c)
            if (not exist):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            
            # Compute intersection distance along ray
            if (t0 > ray.maxt or t1 < ray.mint):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            thit = t0

            if (t0 < ray.mint):
                thit = t1
                if (thit > ray.maxt):
                    if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                    else : return sh_name, r, None, False, None, None, None, None, False

            # Compute sphere hit position and $\phi$
            phit = ray(thit)
            if (phit.x == 0 and phit.y == 0): phit.x = 1e-5 * self.alpha
            phi = math.atan2(phit.y, phit.x) # because alpha=beta
            if (phi < 0): phi += TWO_PI

            # Find parametric representation of sphere hit
            u = phi / TWO_PI
            theta = math.acos(clamp(phit.z / self.gamma, -1, 1))
            v = 1 - (theta / math.pi)

            # Compute sphere dpdu and dpdv
            zradius = math.sqrt(phit.x*phit.x + phit.y*phit.y)
            invzradius = 1 / zradius
            cosphi = phit.x * invzradius
            sinphi = phit.y * invzradius
            fac = -math.pi*(self.alpha/self.gamma)*phit.z
            dpdu = Vector(-TWO_PI*phit.y, TWO_PI*phit.x, 0.)
            dpdv = Vector(fac*cosphi, fac*sinphi, math.pi*self.gamma*math.sin(theta))

            out = sh_name, r, thit, True, u, v, self.oTw(dpdu).to_numpy(), self.oTw(dpdv).to_numpy(), False
            if ds_output : return get_intersect_dataset(*out)
            else : return out
    
    def area(self):
        """
        compute the spheroid area

        .. warning::
            `the scale transformation is not considered for the area calculation!`
        """
        if (self.gamma < self.alpha): # oblate spheroid
            e = math.sqrt(1 - (self.gamma2/self.alpha2))
            area = TWO_PI*self.alpha2 + math.pi*(self.gamma2/e)*math.log((1+e)/(1-e))
        elif (self.gamma > self.alpha): # prolate
            e = math.sqrt(1 - (self.alpha2/self.gamma2))
            area = TWO_PI*self.alpha2*(1 + (self.gamma/(self.alpha*e))*math.asin(e))
        else: # sphere
            area = 4.*math.pi*self.alpha2
        return area
    
    def to_trianglemesh(self, reso_theta=None, reso_phi=None):
        """
        Convert the spheroid to a triangle mesh

        Parameters
        ----------
        reso_theta : int, optional
            The number of lines around the polar theta angle, minimum accepted value is 3
        reso_phi : int, optional
            The number of lines around the azimuth phi angle, minimum accepted value is 3

        Returns
        -------
        mesh : TriangleMesh
            The spheroid converted to a triangle mesh
        """
        rescale_xyz = get_scale_tf(Vector(self.alpha, self.alpha, self.gamma))
        msh = create_sphere_trianglemesh(radius=1, reso_theta=reso_theta, reso_phi=reso_phi)
        vertices_t = np.zeros((msh.nvertices,3))
        for iver in range (0, msh.nvertices):
            vertices_t[iver,:] = rescale_xyz(Point(msh.vertices[iver,:])).to_numpy()
        return TriangleMesh(vertices_t, msh.faces, oTw=self.oTw, wTo=self.wTo)
    
    def plot(self, **kwargs):
        """
        Plot the spheroid

        - The spheroid is first converted to a triangle mesh then the TriangleMesh 
          plot method is used

        Parameters
        ----------
        **kwargs
            The keyword arguments are passed on to the TriangleMesh plot method
        """
        return self.to_trianglemesh().plot(**kwargs)
    

class Disk(Shape):
    '''
    Creation of the class Disk

    Parameters
    ----------
    radius : float
        The disk radius
    inner_radius : float, optional
        The inner radius (case of annulus)
    phi_max : float, optional
        The maximum phi value in degrees of the disk/annulus, where phi is between 0 and 360°
    z_height : float, optional
        the disk height along the z axis
    oTw : Transform, optional
        From object to world space or the transformation applied to the spheroid
    wTo : Transform, optional
        From world to object space or the in inverse transformation applied to the spheroid
    
    Notes
    -----
    Even if z_height is given, the origin for rotation transformation do not change.
    For exemple: z_height=5 then we apply a rotation of 90 degrees around the y axis, the disk
    we be rotated from (0.,0.,0.), meaning the disk we be moved from position (0.,0.,5.) to
    (5.,0.,0.).
    '''
    def __init__(self, radius, inner_radius=0., phi_max=360., z_height=0., oTw=None, wTo=None):
        if wTo is None and oTw is None:
            wTo = Transform()
            oTw = Transform()
        if (phi_max < 0. or phi_max > 360.):
            raise ValueError ('The value of the parameter phi_max must in the range: [0, 360]')
        elif (wTo is None and isinstance(oTw, Transform)): wTo = oTw.inverse()
        elif (isinstance(wTo, Transform) and oTw is None): oTw = wTo.inverse()
        if (inner_radius >= radius): raise NameError ('The parameter inner_radius must be < to radius')
        if (not np.isscalar(radius)): raise ValueError('The parameters radius must be a scalar')
        if (not np.isscalar(inner_radius)): raise ValueError('The parameters inner_radius must be a scalar')
        if (not np.isscalar(phi_max)): raise ValueError('The parameters phi_max must be a scalar')
        if (not np.isscalar(z_height)): raise ValueError('The parameters z_height must be a scalar')
        Shape.__init__(self, ObjectToWorld = oTw, WorldToObject = wTo)
        self.radius = radius
        self.inner_radius = inner_radius
        self.phi_max = phi_max
        self.z_height = z_height

    def is_intersection_t(self, r):
        """
        Test if a ray/set of rays intersects the disk

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        thit : float | 1-D ndarray
            The t ray variable(s) for its first intersection at the shape surface
        is_intersection : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> r1 = gc.Ray(gc.Point(1.2,0.,10.), gc.Vector(0.,0.,-1.))
        >>> r2 = gc.Ray(gc.Point(0.2,0.,10.), gc.Vector(0.,0.,-1.))
        >>> r3 = gc.Ray(gc.Point(1.6,0.,10.), gc.Vector(0.,0.,-1.))
        >>> annulus = gc.Disk(radius=1.5, inner_radius=0.8)
        >>> annulus.is_intersection_t(r1) # hit point is between the inner radius and radius
        (10.0, True)
        >>> annulus.is_intersection_t(r2) # the ray passes through the annulus hole, no intersection
        (None, False)
        >>> annulus.is_intersection_t(r3) # the ray passes outside, no intersection
        (None, False)
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # no intersection in the case the ray is parallel to the disk's plane
                c1 = ray.d.z == 0
                thit = (self.z_height - ray.o.z) / ray.d.z
                c2 = np.logical_or(thit <= 0, thit >= ray.maxt)

                # get the intersection point, and distance between disk center and the intersection
                phit = ray(thit)
                hit_radius2 = phit.x*phit.x + phit.y*phit.y

                # if the hit point is outside the disk then no intersection
                c3 = hit_radius2 > self.radius*self.radius
            
                # annulus case
                # check that the hit point is not in the annulus hole
                c4 = hit_radius2 < self.inner_radius*self.inner_radius

                # partial disk/annulus case
                # check phi value to see if the hit point is inside the partial disk/annulus
                phi = np.arctan2(phit.y, phit.x)
                phi_max_rad = math.radians(self.phi_max)
                phi[phi < 0.] += TWO_PI
                c5 = phi > phi_max_rad

                c6 = np.logical_or.reduce((c1, c2, c3, c4, c5))
                is_intersection[c6] = False
                thit[c6] = None

                return thit, is_intersection
        else:
            # no intersection in the case the ray is parallel to the disk's plane
            if (ray.d.z == 0): return None, False
            thit = (self.z_height - ray.o.z) / ray.d.z
            if (thit <= 0 or thit >= ray.maxt): return None, False

            # get the intersection point, and distance between disk center and the intersection
            phit = ray(thit)
            hit_radius2 = phit.x*phit.x + phit.y*phit.y

            # if the hit point is outside the disk then no intersection
            if (hit_radius2 > self.radius*self.radius): return None, False

            # annulus case
            # check that the hit point is not in the annulus hole
            if (self.inner_radius > 0.):
                if (hit_radius2 < self.inner_radius*self.inner_radius): return None, False
            
            # partial disk/annulus case
            # check phi value to see if the hit point is inside the partial disk/annulus
            if (self.phi_max < 360.):
                phi = math.atan2(phit.y, phit.x)
                if (phi < 0.): phi += TWO_PI
                if (phi > math.radians(self.phi_max)): return None, False
            
            return thit, True
    
    def is_intersection(self, r):
        """
        Test if a ray/set of rays intersects the disk

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test

        Returns
        -------
        is_intersection : bool | 1-D ndarray
            If there is an intersection -> True, else False

        Examples
        --------
        >>> import geoclide as gc
        >>> r1 = gc.Ray(gc.Point(1.2,0.,10.), gc.Vector(0.,0.,-1.))
        >>> r2 = gc.Ray(gc.Point(0.2,0.,10.), gc.Vector(0.,0.,-1.))
        >>> r3 = gc.Ray(gc.Point(1.6,0.,10.), gc.Vector(0.,0.,-1.))
        >>> annulus = gc.Disk(radius=1.5, inner_radius=0.8)
        >>> annulus.is_intersection(r1) # hit point is between the inner radius and radius
        True
        >>> annulus.is_intersection(r2) # the ray passes through the annulus hole, no intersection
        False
        >>> annulus.is_intersection(r3) # the ray passes outside, no intersection
        False
        """
        _, is_intersection = self.is_intersection_t(r)
        return is_intersection

    def intersect(self, r, ds_output=True):
        """
        Test if a ray/set of rays intersects the disk

        Parameters
        ----------
        r : Ray
            The ray(s) to use for the intersection test(s)
        ds_output : Bool, optional
            If True the output is a dataset, else -> a tuple with intersection information variables

        Returns
        -------
        out : xr.Dataset | tuple
            Look-up table with the intersection information if ds_output is True, 
            else returns a tuple. The tuple is ready to be an input for the function 
            geoclide.shapes.get_intersect_dataset

        Examples
        --------
        >>> import geoclide as gc
        >>> r1 = gc.Ray(gc.Point(1.2,0.,10.), gc.Vector(0.,0.,-1.))
        >>> annulus = gc.Disk(radius=1.5, inner_radius=0.8)
        >>> annulus.intersect(r1) # hit point is between the inner radius and radius
        <xarray.Dataset> Size: 209B
        Dimensions:          (xyz: 3)
        Coordinates:
        * xyz              (xyz) int64 24B 0 1 2
        Data variables:
            o                (xyz) float64 24B 1.2 0.0 10.0
            d                (xyz) float64 24B 0.0 0.0 -1.0
            mint             int64 8B 0
            maxt             float64 8B inf
            is_intersection  bool 1B True
            thit             float64 8B 10.0
            u                float64 8B 0.0
            v                float64 8B 0.4286
            phit             (xyz) float64 24B 1.2 0.0 0.0
            nhit             (xyz) float64 24B 0.0 0.0 1.0
            dpdu             (xyz) float64 24B 0.0 7.54 0.0
            dpdv             (xyz) float64 24B -0.7 -0.0 -0.0
        """
        if not isinstance(r, Ray): raise ValueError('The given parameter must be a Ray')
        sh_name = self.__class__.__name__
        is_r_arr = isinstance(r.o.x, np.ndarray)
        if is_r_arr: nrays = len(r.o.x)
        ray = Ray(r)
        ray.o = self.wTo(r.o)
        ray.d = self.wTo(r.d)

        if is_r_arr:
            with np.errstate(divide='ignore', invalid='ignore'):
                is_intersection = np.full(nrays, True, dtype=bool)

                # no intersection in the case the ray is parallel to the disk's plane
                c1 = ray.d.z == 0
                thit = (self.z_height - ray.o.z) / ray.d.z
                c2 = np.logical_or(thit <= 0, thit >= ray.maxt)

                # get the intersection point, and distance between disk center and the intersection
                phit = ray(thit)
                hit_radius2 = phit.x*phit.x + phit.y*phit.y

                # if the hit point is outside the disk then no intersection
                c3 = hit_radius2 > self.radius*self.radius
            
                # annulus case
                # check that the hit point is not in the annulus hole
                c4 = hit_radius2 < self.inner_radius*self.inner_radius

                # partial disk/annulus case
                # check phi value to see if the hit point is inside the partial disk/annulus
                phi = np.arctan2(phit.y, phit.x)
                phi_max_rad = math.radians(self.phi_max)
                phi[phi < 0.] += TWO_PI
                c5 = phi > phi_max_rad

                # get the parameteric representation
                u = phi / phi_max_rad
                hit_radius = np.sqrt(hit_radius2)
                v = (self.radius-hit_radius) / (self.radius-self.inner_radius)
                zeros = np.zeros(nrays, dtype=np.float64)
                dpdu = Vector(-phi_max_rad*phit.y, phi_max_rad*phit.x, zeros)
                dpdv = Vector(phit.x, phit.y, zeros) * ( (self.inner_radius-self.radius)/hit_radius )

                c6 = np.logical_or.reduce((c1, c2, c3, c4, c5))
                is_intersection[c6] = False
                thit[c6] = None

                out = sh_name, r, thit, is_intersection, u, v, self.oTw(dpdu).to_numpy(), \
                    self.oTw(dpdv).to_numpy(), False
                if ds_output : return get_intersect_dataset(*out)
                else : return out
        else:
            # no intersection in the case the ray is parallel to the disk's plane
            if (ray.d.z == 0):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            thit = (self.z_height - ray.o.z) / ray.d.z
            if (thit <= 0 or thit >= ray.maxt):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False

            # get the intersection point, and distance between disk center and the intersection
            phit = ray(thit)
            hit_radius2 = phit.x*phit.x + phit.y*phit.y

            # if the hit point is outside the disk then no intersection
            if (hit_radius2 > self.radius*self.radius):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False

            # annulus case
            # check that the hit point is not in the annulus hole
            if (hit_radius2 < self.inner_radius*self.inner_radius):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False
            
            # partial disk/annulus case
            # check phi value to see if the hit point is inside the partial disk/annulus
            phi = math.atan2(phit.y, phit.x)
            phi_max_rad = math.radians(self.phi_max)
            if (phi < 0.): phi += TWO_PI
            if (phi > phi_max_rad):
                if ds_output : return get_intersect_dataset(sh_name, r, None, False, None, None, None, None, False)
                else : return sh_name, r, None, False, None, None, None, None, False

            # get the parameteric representation
            u = phi / phi_max_rad
            hit_radius = math.sqrt(hit_radius2)
            v = (self.radius-hit_radius) / (self.radius-self.inner_radius)
            dpdu = Vector(-phi_max_rad*phit.y, phi_max_rad*phit.x, 0.)
            dpdv = Vector(phit.x, phit.y, 0.) * ( (self.inner_radius-self.radius)/hit_radius )

            out = sh_name, r, thit, True, u, v, self.oTw(dpdu).to_numpy(), self.oTw(dpdv).to_numpy(), False
            if ds_output : return get_intersect_dataset(*out)
            else : return out
    
    def area(self):
        """
        compute the disk / annulus area

        .. warning::
            `the scale transformation is not considered for the area calculation!`
        """
        return 0.5*math.radians(self.phi_max)*(self.radius*self.radius - self.inner_radius*self.inner_radius)

    def to_trianglemesh(self, reso=None):
        """
        Convert the disk to a triangle mesh

        Parameters
        ----------
        reso : int, optional
            The number of lines around the polar phi angle, minimum accepted value is 3

        Returns
        -------
        mesh : TriangleMesh
            The disk converted to a triangle mesh
        """
        return create_disk_trianglemesh(self.radius, self.inner_radius, reso, self.phi_max,
                                        self.z_height, self.oTw, self.wTo)
    
    def plot(self, **kwargs):
        """
        Plot the disk

        - The disk is first converted to a triangle mesh then the TriangleMesh 
          plot method is used

        Parameters
        ----------
        **kwargs
            The keyword arguments are passed on to the TriangleMesh plot method
        """
        return self.to_trianglemesh().plot(**kwargs)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import geoclide as gc


def test_clamp():
    assert (gc.clamp(4., val_min=5., val_max=11.) == 5.)


def test_quadratic():
    a = 2.
    b = -5.
    c = 0.
    assert (gc.quadratic(a, b, c) == (True, 0.0, 2.5))